# %%

# TODO: make this demo less shitty

import os

import pytest
import torch

import rust_circuit as rc
import rust_circuit.module_library as mod_l
from rust_circuit.model_rewrites import configure_transformer

os.environ["RUST_BACKTRACE"] = "1"

# %%

params = mod_l.TransformerParams(mod_l.TransformerBlockParams(attn_bias=True, mlp_output_bias=True), num_layers=2)
new_circ, _, _ = params.garbage_call(num_heads=2)  # 2 heads to avoid too much clutter
new_circ = new_circ.update("t.bind_w", lambda x: configure_transformer(x, use_pull_up_bias=False))
new_circ.print()

# %%

x = new_circ.get_unique("b1")
print()
x.print()
print()
# push_down_modules currently doesn't support unknown sizes well, and will sometimes error. TODO fix
with pytest.raises(rc.RearrangeSpecIntOnlyInOutputWithoutSizeError):
    pushed = rc.ModulePusher()(x, traversal=rc.new_traversal(term_early_at={"m", "ln", "a.on_inp"}))
# pushed.print()
# rep_push = new_circ.update("b1", lambda _: pushed)
# torch.testing.assert_close(rep_push.evaluate(), new_circ.evaluate())

# %%

# crazy pushing!
for flatten_modules in [False, True]:
    for end_depth in range(8):
        full_push = rc.ModulePusher(flatten_modules=flatten_modules)(
            new_circ, traversal=rc.new_traversal(end_depth=end_depth), skip_module=~rc.Matcher(new_circ)
        )
        torch.testing.assert_close(full_push.evaluate(), new_circ.evaluate())

# %%

new_circ_extra_batch = rc.Expander(
    ("t.inp.tok_embeds", lambda x: rc.Array.randn(2, *x.shape, device_dtype=rc.TorchDeviceDtypeOp(dtype="float64")))
)(new_circ)

# more crazy pushing!
for flatten_modules in [False, True]:
    for end_depth in range(8):
        full_push = rc.ModulePusher(flatten_modules=flatten_modules)(
            new_circ_extra_batch, traversal=rc.new_traversal(end_depth=end_depth), skip_module=~rc.Matcher(new_circ)
        )
        torch.testing.assert_close(full_push.evaluate(), new_circ_extra_batch.evaluate())

# %%

# partial substitution with just causal mask: first extract then sub
sub_causal_mask = rc.extract_symbols_get(new_circ, "a.mask").substitute()
# very nice
sub_causal_mask.print()
torch.testing.assert_close(sub_causal_mask.evaluate(), new_circ.evaluate())

# %%

# pushing past a overridden symbol isn't supported atm (but could be supported, just a bit complicated)

# same symbol, different value - the inner value is used
s = """
'a' Module
  'add' Add
    'b' Module
      'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
      'b.arg' [] Scalar 1.2 ! 'sym'
    'sym'
  'a.arg' [3] Scalar 1.3 ! 'sym'
"""
circ = rc.Parser()(s)
with pytest.raises(rc.PushDownModulePushingPastModuleWhichOverridesSymError):
    rc.ModulePusher().push_down_modules(circ, traversal=rc.new_traversal(), skip_module="b")

# if you don't skip, this is totally valid
torch.testing.assert_close(
    rc.ModulePusher().push_down_modules(circ, traversal=rc.new_traversal()).evaluate(), circ.evaluate()
)

# %%

# more complex case for the above
s = """
'a' Module
  'add' Add
    'b' Module
      'add_inner' Add
        'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
        'sym.outer' [] Symbol 53db5cff-561e-4756-a502-65a26cb2a5f4
      'b.arg' [] Scalar 1.2 ! 'sym'
    'sym'
  'a.arg' [3] Scalar 1.3 ! 'sym'
  'a.outer_arg' [4, 3] Scalar 3.3 ! 'sym.outer'
"""
circ = rc.Parser()(s)
with pytest.raises(rc.PushDownModulePushingPastModuleWhichOverridesSymError):
    rc.ModulePusher().push_down_modules(circ, traversal=rc.new_traversal(), skip_module="b")

torch.testing.assert_close(
    rc.ModulePusher().push_down_modules(circ, traversal=rc.new_traversal(term_early_at="add_inner")).evaluate(),
    circ.evaluate(),
)

rc.ModulePusher().push_down_modules(circ, traversal=rc.new_traversal(term_early_at={"sym", "sym.outer"})).print()
torch.testing.assert_close(
    rc.ModulePusher()
    .push_down_modules(circ, traversal=rc.new_traversal(term_early_at={"sym", "sym.outer"}))
    .evaluate(),
    circ.evaluate(),
)


# %%



# %%

import os
from typing import Optional, Sequence, Tuple

import rust_circuit as rc
from interp.model.model_loading import load_model_info, load_model_mem_cached
from rust_circuit.demos.notebook_testing import NotebookInTesting
from rust_circuit.jax_to_module import get_bound_model
from rust_circuit.module_library import load_transformer_model_string

RRFS_DIR = os.path.expanduser("~/rrfs")
RRFS_INTERP_MODELS_DIR = f"{RRFS_DIR}/interpretability_models_jax/"
os.environ["INTERPRETABILITY_MODELS_DIR"] = os.environ.get(
    "INTERPRETABILITY_MODELS_DIR",
    os.path.expanduser("~/interp_models_new/")
    if os.path.exists(os.path.expanduser("~/interp_models_new/"))
    else RRFS_INTERP_MODELS_DIR,
)
# %%

all_models: Sequence[Tuple[str, str, Optional[str]]]
if NotebookInTesting.currently_in_notebook_test:
    all_models = [
        ("gelu_two_layers_untied", "gelu_2", None),
    ]
else:
    # TODO: resave before merge!
    all_models = [
        ("attention_only_bn_four_layers_nobias_3_fixed", "attention_only_bn_4", None),
        ("attention_only_bn_two_layers_nobias_3_fixed", "attention_only_bn_2", None),
        ("attention_only_two_layers_untied", "attention_only_2", None),
        ("bilinear_bn_four_layers_nobias_3_fixed", "bilinear_bn_4", None),
        ("bilinear_bn_two_layers_nobias_3_fixed", "bilinear_bn_2", None),
        ("bilinear_four_layers_untied", "bilinear_4", None),
        ("bilinear_two_layers_untied", "bilinear_2", None),
        ("gelu_one_layer_untied", "gelu_1", None),
        ("gelu_two_layers_untied", "gelu_2", None),
        ("gelu_twelve_layers", "gelu_12_tied", "gpt2-small"),
        ("gelu_twenty_four_layers", "gelu_24_tied", "gpt2-medium"),
        ("jun9_paren_balancer", "jun9_paren_balancer", None),
        ("paren_balancer_backdoor_nov_16", "paren_balancer_backdoor_nov_16", None)
        # ("gelu_36_layers", "gelu_36", "gpt2-large"),
        # ("gelu_48_layers", "gelu_48", "gpt2-xlarge"),
    ]

for (orig_model_id, new_model_id, aka) in all_models:
    model_info, _ = load_model_info(orig_model_id)
    model, params, _ = load_model_mem_cached(orig_model_id)
    model.norm_type
    bound_model, both_embeds, info, _ = get_bound_model(model.bind(params), model_class=model_info["model_class"])
    s = info.dump_model_string(*both_embeds, bound_model)
    s += f"\n# originally {orig_model_id}"
    if aka is not None:
        s += f" (aka {aka})"
    s += "\n"
    with open(f"/tmp/{new_model_id}.circ", "w") as f:
        f.write(s)

# %%
# Can be slow!
# Helps if you've cleared ~/tensors_by_hash_cache before running this notebook
rc.sync_all_unsynced_tensors()

# %%

# for running models, see interp/circuit/interop_rust/test_jax_to_module.py
out, tok, extra_args = load_transformer_model_string(open("/tmp/gelu_2.circ").read())
# out, tok, extra_args = load_model_id("gelu_2")
print(extra_args)
out["t.logits"].print()
tok



from typing import Callable, Iterable, Optional, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def it(x: Optional[T]) -> Iterable[T]:
    if x is not None:
        yield x


def map(x: Optional[T], f: Callable[[T], U]) -> Optional[U]:
    if x is not None:
        return f(x)
    return None


def unwrap(x: Optional[T], error: str = "") -> T:
    assert x is not None, error
    return x


def unwrap_or(x: Optional[T], default: T) -> T:
    if x is None:
        return default
    return x


def unwrap_or_else(x: Optional[T], default: Callable[[], T]) -> T:
    if x is None:
        return default()
    return x


def or_y(x: Optional[T], y: Optional[T]) -> Optional[T]:
    if x is None:
        return y
    return x



import pytest
import torch

import rust_circuit as rc


def test_err_on_param():
    with pytest.raises(TypeError):
        rc.Array(torch.nn.Parameter(torch.randn(3)))
    with pytest.raises(TypeError):
        rc.hash_tensor(torch.nn.Parameter(torch.randn(3)))
    rc.Array(torch.nn.Parameter(torch.randn(3)).clone())
    rc.hash_tensor(torch.nn.Parameter(torch.randn(3)).clone())



import re
from enum import Enum
from typing import Literal, Optional, Union

import rust_circuit.algebric_rewrite as algrw
import rust_circuit.optional as op

from . import _rust as rc

# from functools import total_ordering


def push_down_transformer_namer(circ: rc.Circuit, _, mods: list[rc.Module], _num):
    if re.match(r"([amb]\.)|([am]$)|(ln$)", circ.name):
        if circ.name == "ln":
            name = mods[0].name
        else:
            name = circ.name
        split_first, _, split_rest = name.partition(".")
        suffix = f".{split_rest}" if "." in name else ""
        num: Optional[int] = None
        for x in mods:
            get_name = re.match(r"b(\d+)$", x.name)
            if get_name is None:
                continue
            num = int(get_name.group(1))
            break

        # this is need to handle final.ln case correctly
        num_str = str(num) if isinstance(num, int) else ""

        return f"{split_first}{num_str}{suffix}"

    return circ.name


# total ordering doesn't type check, sad
# @total_ordering
class PushDownTransformerTo(Enum):
    BLOCK = 0
    ATTN_MLP = 1
    ATTN_MLP_NORM = 2
    ATTN_HEAD_MLP_NORM = 3

    # def __lt__(self, other):
    #     if self.__class__ is other.__class__:
    #         return self.value < other.value
    #     return NotImplemented


To = PushDownTransformerTo


def strip_arr(c: rc.Circuit):
    """strip _arr suffix from array names"""
    return c.update(
        rc.Array & rc.Matcher(lambda x: x.name.endswith("_arr")), lambda x: x.rename(x.name.removesuffix("_arr"))
    )


def pull_up_bias(c: rc.Circuit):
    """
    pull up attention/mlp bias passed module binding sites

    useful because the bias isn't really strongly associated with the attention/mlp

    Might be nice for flattening out the residual stream
    """

    def namer(x: rc.Circuit, _, mods: list[rc.Module], _num):
        if len(mods) == 0:
            return x.name
        norm_call = ["a.norm_call", "m.norm_call"]
        block_call = ["b.a", "b.m"]
        if x.name in ["a", "m"]:
            if mods[-1].name in block_call:
                assert len(mods) in [1, 2]
                if len(mods) == 2:
                    assert mods[0].name in norm_call
                return f"b.{x.name}"
            elif len(mods) == 1:
                assert mods[0].name in norm_call
                return f"{x.name}.norm_call"
            else:
                assert False
        return x.name

    out = rc.ModulePusher(namer=namer, flatten_modules=False).push_down_modules(
        c,
        traversal=rc.new_traversal(term_early_at=rc.Matcher.regex(r"^[am]$")),
        skip_module=~(rc.Matcher.regex(r"^[am].norm_call$") | rc.Matcher.regex(r"^b.[am]$")),
    )
    return out.update(rc.Matcher.regex(r"^b.[am]$") & rc.Add, lambda x: x.rename(x.name + ".p_bias"))


def push_down_transformer_weights(
    c: rc.Circuit,
    to: PushDownTransformerTo,
    use_strip_arr: bool = True,
):
    """
    to: How to push down weights (to more granular modules) if at all. The enum is ordered with lower values
    meaning 'less pushing' and higher values meaning 'more pushing'. None means
    keep the weights bound at the top level.

    use_strip_arr: now that weights have been pushed down, names no longer
    confict with symbols so we can remove _arr as a suffix (if desired).
    """
    assert c.name == "t.bind_w"

    skip_on_fuse = ~rc.Matcher(rc.Matcher.regex(r"^b\d+$"), "final.norm", "t.bind_w", rc.Matcher.regex(r"^[am].norm$"))

    term_early_at: list[rc.MatcherIn] = ["ln", "bn"]
    if to == To.BLOCK:
        term_early_at.append("b")
    elif to == To.ATTN_MLP:
        term_early_at.append(rc.Matcher.regex(r"^[am].norm_call$"))
    if to.value <= To.ATTN_MLP_NORM.value:
        term_early_at.append(rc.Matcher.regex(r"^[am](.p_bias)?$"))
    if to.value <= To.ATTN_HEAD_MLP_NORM.value:
        term_early_at.append(rc.Matcher.regex(r"^m(.p_bias)?$"))
        term_early_at.append("a.head.on_inp")

    # TODO: add support for full substitute?

    # we're not doing anything too interesting here - just pushing down some select modules
    out = rc.ModulePusher(namer=push_down_transformer_namer).push_down_modules(
        c,
        traversal=rc.new_traversal(term_early_at=rc.Matcher.any(*term_early_at)),
        skip_module=skip_on_fuse,
    )

    # not currently needed
    # if to >= To.ATTN_MLP:
    #     out = out.update(rc.Matcher.regex(r"^b\d+.[am].set$"), lambda x: x.rename(x.name.removesuffix(".set")))

    if use_strip_arr:
        out = strip_arr(out)

    return out


SplitConfigDict = dict[int, list[tuple[rc.TorchAxisIndex, str]]]
SplitConfigSetDict = Union[set[int], SplitConfigDict]
SplitConfig = Union[Literal["full"], SplitConfigSetDict]

SplitPosConfig = dict[int, list[tuple[Union[slice, int], str]]]


def split_by_head(c: rc.Circuit, split: SplitConfig = "full"):
    """
    requires that weights have been pushed down all the way to heads

    full: split heads on all layers
    set[int]: split heads on these layers
    dict[int, list[...]]: split heads on these layers into this list of groups
        where each group has an index and name
    """

    heads = c.get(rc.Matcher.regex(r"^a\d+.heads$"))
    num_layers = len(heads)
    split_v: SplitConfigSetDict
    if split == "full":
        split_v = set(range(num_layers))
    else:
        split_v = split

    if num_layers == 0:
        return c

    num_heads = list(heads)[0].shape[0]
    split_vv: SplitConfigDict
    if isinstance(split_v, set):
        split_vv = {l: [(h, f"h{h}") for h in range(num_heads)] for l in split_v}
    else:
        split_vv = split_v

    idxed = c
    for i, items in split_vv.items():
        idxs = [idx for idx, _ in items]
        names = [name for _, name in items]
        assert idxed.are_any_found(f"a{i}")
        from .algebric_rewrite import explicit_reduce

        idxed = idxed.update(
            f"a{i}",
            lambda x: explicit_reduce(
                x.cast_einsum(),
                axis=0,
                element_index=0,
                partitioning_idxs=idxs,
                use_axis_name=False,
                partitioning_idx_names=names,
                run_on_index=lambda x, _, s: rc.push_down_index(x, suffix="." + s),
                use_dot=True,
            ),
        )
    return rc.simp(idxed).update(rc.Matcher.regex(r"^a\d+.heads."), lambda x: x.rename(x.name.replace(".heads.", ".")))


def pull_up_head_split(c: rc.Circuit, to_pull_up: Optional[set[int]] = None):
    """
    pulls up split heads past modules: this results in the layernorm and input
    being bound for each head individually.

    This might be nice for flatten_res.
    """

    def namer(x: rc.Circuit, _, mods: list[rc.Module], _num):
        block_call = r"b\d+\.a$"
        norm_call = r"a\d+\.norm_call$"
        if len(mods) > 0 and re.match(rf"{norm_call}|{block_call}", mods[0].name):
            assert re.match(r"^a\d+", x.name), x.name
            if re.match(block_call, mods[-1].name):
                assert len(mods) in [1, 2]
                if len(mods) == 2:
                    assert re.match(norm_call, mods[0].name)
                attn, _, head_rest = x.name.partition(".")
                num = int(attn[1:])
                return f"b{num}.a.{head_rest}"
            elif len(mods) == 1:
                assert re.match(norm_call, mods[0].name)
                return f"{x.name}.norm_call"
            assert False
        return x.name

    num_layers = len(c.get(rc.Matcher.regex(r"^a\d+$")))
    to_pull_up_v = op.unwrap_or(to_pull_up, set(range(num_layers)))
    out = c
    for i in to_pull_up_v:
        out = rc.ModulePusher(namer=namer, flatten_modules=False).push_down_modules(
            out,
            traversal=rc.new_traversal(term_early_at=~rc.Matcher.match_any_found(f"a{i}")).chain(
                rc.IterativeMatcher.term()
            ),
            skip_module=~(rc.Matcher(f"a{i}.norm_call") | f"b{i}.a"),
        )
    return out


def clear_block_module(c: rc.Circuit):
    """
    substitute out calling blocks on arguments.

    This is only nice to do if you've already pushed weights past blocks.

    This ensures that adding into the residual stream isn't inside a module.
    """
    return c.update(
        rc.Matcher.regex(r"^b\d+.call$"), lambda x: x.cast_module().substitute().rename(x.name.partition(".")[0])
    )


def flatten_res(c: rc.Circuit, flatten_components: bool = False):
    """
    flatten the residual stream

    flatten_components makes flatten operate within mlp/attn components. This
    only matters if you have a bias or attn is separated into heads.
    """
    if flatten_components:
        traversal = rc.new_traversal()
    else:
        traversal = rc.new_traversal(
            term_early_at=rc.Matcher.regex(r"^b\d+.[am](.p_bias)?$") | rc.Matcher.regex("^[am].(.p_bias)?")
        )

    def f(x):
        return rc.add_flatten(x.cast_add(), traversal=traversal)

    return c.update(
        rc.Matcher.regex(r"^b\d+(.resid.\w)?$"), lambda x: f(x) if x.is_add() else x.cast_concat().map_children(f)
    )


def configure_transformer(
    c: rc.Circuit,
    to: Optional[To] = To.ATTN_MLP_NORM,
    use_pull_up_bias: bool = True,
    use_strip_arr: bool = True,
    split_by_head_config: Optional[SplitConfig] = None,
    split_by_pos_config: Optional[SplitPosConfig] = None,
    use_pull_up_head_split: bool = False,
    use_clear_block_module: bool = True,
    use_flatten_res: bool = False,
    flatten_components: bool = False,
    check_valid: bool = True,
    sequence_length: Optional[int] = None,
    is_causal: Optional[bool] = None,
):
    """
    See docs for pull_up_bias, push_down_transformer_weights, split_by_head,
    pull_up_head_split, clear_block_module, and flatten_res.

    This function just subsequently runs these operations (if they are enabled
    via a flag or by having their configuration set to non-None)
    """
    assert c.name == "t.bind_w"

    if check_valid:
        if use_strip_arr:
            assert to is not None, "stripping arr without pushing down results in name collision"
        if split_by_head_config is not None:
            assert to == To.ATTN_HEAD_MLP_NORM, "You can't split by head without pushing down weights to head"
        if split_by_pos_config is not None:
            assert (
                to == To.ATTN_HEAD_MLP_NORM and sequence_length is not None
            ), "You can't split by position without pushing down weights to head and specifying sequence length"
            assert is_causal, "Split by position currently only supports causal models"
        if use_pull_up_head_split:
            assert split_by_head_config is not None, "You can't pull up split heads if heads aren't split!"
        if use_clear_block_module:
            assert (
                to is not None and to.value >= To.ATTN_MLP.value
            ), "clearing block module not supported without pushing down to at least ATTN_MLP"
        if use_flatten_res:
            assert (
                use_clear_block_module
            ), "if you want to flatten the residual stream, you'll need to use_clear_block_module"
        if flatten_components:
            assert use_flatten_res, "setting flatten_components has no effect without use_flatten_res"

    if use_pull_up_bias:
        c = pull_up_bias(c)
    if to is not None:
        c = push_down_transformer_weights(c, to, use_strip_arr=use_strip_arr)
    if split_by_head_config is not None:
        c = split_by_head(c, split_by_head_config)
    if use_pull_up_head_split:
        to_pull_up: Optional[set[int]] = None
        if isinstance(split_by_head_config, set):
            to_pull_up = split_by_head_config
        elif isinstance(split_by_head_config, dict):
            to_pull_up = set(split_by_head_config.keys())
        c = pull_up_head_split(c, to_pull_up=to_pull_up)
    if split_by_pos_config:
        assert sequence_length is not None
        c = split_by_pos(c, split_by_pos_config, sequence_length)
    if use_clear_block_module:
        c = clear_block_module(c)
    if use_flatten_res:
        c = flatten_res(c, flatten_components=flatten_components)
        if split_by_pos_config and use_clear_block_module:

            def try_(f, x):
                return f(x) or x

            c = c.update({"b.input idx idx"}, lambda x: op.unwrap(rc.index_fuse(x.cast_index())))
            c = c.update(
                {"b.input idx"},
                lambda x: try_(rc.index_elim_identity, try_(rc.index_concat_drop_unreached, x)),
            )

    return c


def batch_mul(
    x: rc.Einsum,
    operand_idx: int,
    retained_axes: tuple[int, ...],
    insert_at_idx: Optional[int] = 0,  # None: end because -0 doesn't exist
    assert_not_present: bool = True,
    assert_not_diag: bool = True,
):
    einsum_axes = [x.all_input_axes()[operand_idx][i] for i in retained_axes]
    if assert_not_diag:
        assert len(einsum_axes) == len(set(einsum_axes))

    if assert_not_present:
        assert len(set(einsum_axes).intersection(x.out_axes)) == 0
    insert_at_idx_v = op.unwrap_or(insert_at_idx, len(x.out_axes))
    return x.evolve(out_axes=x.out_axes[:insert_at_idx_v] + tuple(einsum_axes) + x.out_axes[insert_at_idx_v:])


# po = rc.PrintOptions(
#     # traversal=rc.restrict(m, True),
#     colorer=rc.PrintOptions.type_colorer(),
#     bijection=False,
#     shape_only_when_necessary=False,arrows=True
# )


def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
    """
    requires that weights have been pushed down all the way to heads, sequence length is specified as len_ (sequence_length in configure_transformer), and the model has causal attention (is_causal in configure_transformer)

    {layer: [(slice, name), (int, name2), ...], layer2: ...}: split these layers by position into this list of groups
    """

    def push_down(x):
        return rc.push_down_index(
            x,
            rc.restrict(
                {rc.Symbol, rc.Index},
                True,
            ),
            allow_partial_pushdown=True,
            into_module_spec=~rc.Matcher({"a.q"}),
        )

    def update_layer(ca, splits: list[Union[int, slice]], split_names: list[str]):
        mask = ca.get_unique("a.mask")
        mask = rc.Concat(
            *(
                rc.Concat(
                    rc.Index(mask, (slice(ss, se), slice(0, se))),
                    rc.Scalar(0.0, (se - ss, len_ - se)),
                    axis=1,
                    name=f"a.mask_causal_at_{n}",
                )
                for (ss, se, n) in [
                    (x.start, x.stop, n) if isinstance(x, slice) else (x, x + 1, n)
                    for (x, n) in zip(splits, split_names)
                ]
            ),
            axis=0,
            name="a.mask causal",
        )

        ca = rc.Expander(
            ("a.mask", lambda _: mask), ("b.input", lambda x: rc.Index(x, (slice(0, len_), (slice(None, None)))))
        )(ca)
        ca = algrw.split_to_concat(
            ca, 0, splits, split_names, run_on_index=lambda x, *_: push_down(x).rename(f"{x.name}")
        ).rename(ca.name + "_concat_pos")

        ca = ca.update(
            "a.mask causal", lambda x: op.unwrap(rc.concat_elim_identity(x.cast_concat()))
        )  # left by push_down_index
        ca = ca.update("a.attn_scores", rc.deep_pull_concat_messy)

        def upd_softmax(x: rc.Circuit) -> rc.Concat:
            assert isinstance(x, rc.GeneralFunction)
            assert isinstance(x.spec, rc.GeneralFunctionSimpleSpec)
            assert x.spec.name == "softmax"
            concat = x.children[0]
            assert isinstance(concat, rc.Concat)
            assert concat.num_children == 2  # before & self, after
            assert concat.axis == 1
            return rc.Concat(
                rc.softmax(concat.children[0], name="a.attn_probs").rename("a.attn_probs"),
                rc.Scalar(0.0, concat.children[1].shape, name="a.attn_probs_zero"),
                axis=1,
                name="a.attn_probs",
            )

        ca = ca.update("a.attn_probs", upd_softmax)
        ca = ca.update(
            rc.Regex("^a.mask_causal_at_"), lambda x: rc.concat_drop_size_zero(x.cast_concat()) or x
        )  # fixup last

        ca = ca.update("(a.attn_scores) idx", push_down)

        ca = ca.update("a.comb_v", lambda x: op.unwrap(rc.pull_concat_once(x)))

        ca = ca.update(rc.Matcher("a.attn_probs") & rc.Add, lambda x: x.children[0])  # other is zero
        # TODO: use compiler_simp to prove it, needs more rewrites
        # ca = ca.update(rc.Matcher('a.attn_probs') & rc.Add, lambda x: x.map_children_enumerate(lambda ix, c: rc.compiler_simp(c) if ix == 1 else c)) # other is zero

        ca = ca.update({"(a.v) idx", "(a.v_p_bias) idx"}, push_down)
        ca = ca.update(
            rc.Matcher("a.attn_scores_raw").chain(rc.restrict(rc.Index, start_depth=1, end_depth=2)), push_down
        )

        # ca = ca.update(
        #     rc.Matcher("a.v").chain(rc.restrict("a.v.input set_shape", end_depth=2))
        #     | rc.Matcher("a.k").chain(rc.restrict("a.k.input set_shape", end_depth=2)),
        #     lambda a: rc.Index(a, [slice(0, a.shape[0]), slice(None, None)]),
        # )  # fixup last

        def split_idx(ix):
            ix = ix.cast_index()
            if ix.idx[0].start == 0:
                return ix
            return rc.Index(rc.Index(ix.node, (slice(0, ix.idx[0].stop), slice(None, None))), ix.idx)

        ca = ca.update({"a.q.input idx", "p.input idx"}, split_idx)

        ca = algrw.push_down_indexes_through_modules(
            ca,
            {
                "a.v.input idx",
                "a.k.input idx",
                "a.q.input idx",
                "p.input idx",
                "p.input",
            },
            push_down=push_down,
        )
        return ca

    for (l, c) in config.items():
        if isinstance(c[0][0], str):
            raise Exception(
                "split_by_pos config was changed to match split_by_head, was (name, idxs), now is (idxs, name) at innermost level"
            )
        names = [n for _, n in c]
        splits = [s for s, _ in c]
        circ = circ.update(
            rc.IterativeMatcher(f"b{l}.call").spec_circuit_matcher(), lambda x: update_layer(x, splits, names)
        )

    return circ



from typing import Optional

import jax
import jax.numpy as jnp
import pytest
import torch

import rust_circuit as rc
import rust_circuit.optional as op
from interp.circuit.circuit_models import get_trained_gpt
from interp.circuit.circuit_utils import cast_circuit
from interp.circuit.constant import ArrayConstant
from interp.circuit.get_update_node import FunctionIterativeNodeMatcher, NameMatcher, NodeGetter
from interp.circuit.test_circuit_models import CONFIGS, GptTestConfig, get_gpt_test
from interp.model.gpt_model import Gpt, module_config_dict
from interp.tools.data_loading import get_val_seqs
from interp.tools.variable_dict import variable_dict_map
from rust_circuit.jax_to_module import from_converted, get_bound_model
from rust_circuit.module_library import get_model_path, load_model_id

from .test_models import evaluate_circ


def pos_mask_to_inp_mask(pos_mask: Optional[torch.Tensor]):
    return op.map(pos_mask, lambda pos_mask: rc.Array(1 - pos_mask, name="a.pos_mask"))


@pytest.mark.parametrize("config", CONFIGS)
def test_model_gpt(
    config: GptTestConfig,
    seed=2833333,
    head_size=5,
    seq_len=3,
    num_heads=4,
    batch_size=2,
    num_layers=2,
    vocab_size=27,
):
    jax.config.update("jax_enable_x64", True)
    # TODO: we could maybe cache identical tests instead of removing so we don't unintentionally remove coverage
    if (
        config.use_fits
        or config.norm_type not in {"none", "layer_norm", "batch_norm_fixed"}
        or not config.use_batch
        or config.output_type != "probs"  # we fix output type, so varying this from default does nothing
        or config.use_onehot_tokens
        or config.softmax_type != "softmax"
    ):
        return

    config.output_type = "logits"

    circuit, _, model, params, _, pos_mask = get_gpt_test(
        key=jax.random.PRNGKey(seed),
        head_size=head_size,
        seq_len=seq_len,
        num_heads=num_heads,
        batch_size=batch_size,
        num_layers=num_layers,
        vocab_size=vocab_size,
        config=config,
    )
    circuit = cast_circuit(circuit, dtype=torch.float64)
    model = Gpt(**{**module_config_dict(model), **dict(dtype=jnp.float64)})
    params = variable_dict_map(params, lambda x: x.astype(jnp.float64))

    def get(s: str):
        return NodeGetter(FunctionIterativeNodeMatcher(NameMatcher(s))).get_unique_c(circuit)

    expected_val = evaluate_circ(circuit)
    assert expected_val.shape == circuit.shape

    def get_arr(s: str):
        return ArrayConstant.unwrap(get(s))

    model_b = model.bind(params)
    new_circ, (_, pos_embed_weights), info, _ = get_bound_model(model_b)
    if pos_mask is not None:
        pos_mask_v = from_converted(pos_mask).value
    else:
        pos_mask_v = None
    new_circ_bound = info.bind_to_input(
        new_circ,
        inp_tok_embeds=rc.Array(get_arr("tok_embeds").value.double(), name="t.inp.tok_embeds"),
        pos_embed_weights=pos_embed_weights,
        inp_mask=pos_mask_to_inp_mask(pos_mask_v),
    )
    actual_val = new_circ_bound.evaluate()
    assert actual_val.shape == new_circ_bound.shape

    torch.testing.assert_close(actual_val, expected_val, rtol=1e-5, atol=1e-5)

    jax.config.update("jax_enable_x64", False)


def raw_test_load_circ_model(model_id: str, seq_len: int = 64, batch_size: int = 4):
    s = open(get_model_path(model_id)).read()
    ls = [x.removeprefix("# originally ") for x in s.split("\n") if x.startswith("# originally ")]
    assert len(ls) == 1
    old_model_id = ls[0].split(" ")[0]

    toks = torch.tensor(get_val_seqs(train=True, n_files=1, files_start=0, max_size=seq_len + 1)[:batch_size, :seq_len])

    new_circs, _, info = load_model_id(model_id)
    new_circ = new_circs["t.bind_w"]
    device_dtype = rc.TorchDeviceDtypeOp(dtype="float64", device="cpu")
    new_circ = rc.cast_circuit(new_circ, device_dtype)
    pos_embeds = rc.cast_circuit(new_circs["t.w.pos_embeds"], device_dtype)

    inp_tok_embeds = new_circs["t.w.tok_embeds"].cast_array().value.double().cpu()[toks]
    new_circ = info.bind_to_input(new_circ, rc.Array(inp_tok_embeds, name="t.inp.tok_embeds"), pos_embeds)
    assert new_circ.is_explicitly_computable

    circ, _, _, _, _ = get_trained_gpt(
        toks,
        model_id=old_model_id,
        use_poly_fit=False,
        output_type="logits",
        use_batch=True,
    )
    circ = cast_circuit(circ, dtype=torch.float64)

    torch.testing.assert_close(new_circ.evaluate(), evaluate_circ(circ), atol=1e-5, rtol=1e-5)


@pytest.mark.skip
@pytest.mark.parametrize(
    "model_id",
    [
        "attention_only_bn_2",
        "attention_only_bn_4",
        "attention_only_2",
        "bilinear_bn_2",
        "bilinear_bn_4",
        "bilinear_2",
        "bilinear_4",
        "gelu_1",
        "gelu_2",
    ],
)
def test_load_circ_model(model_id: str):
    raw_test_load_circ_model(model_id, seq_len=16, batch_size=3)


@pytest.mark.skip
@pytest.mark.parametrize("model_id", ["gelu_12", "gelu_24"])
def test_load_circ_model_big(model_id: str):
    raw_test_load_circ_model(model_id, seq_len=4, batch_size=2)



import functools
import hashlib
import weakref
from typing import *

if TYPE_CHECKING:
    from interp.circuit.circuit import Circuit, Shape

import torch

import rust_circuit.optional as op
from rust_circuit.py_utils import assert_never

from . import _rust as rc
from ._rust import Add as rAdd
from ._rust import Array as rArray
from ._rust import Circuit as rCircuit
from ._rust import Concat as rConcat
from ._rust import Cumulant as rCumulant
from ._rust import DiscreteVar as rDiscreteVar
from ._rust import Einsum as rEinsum
from ._rust import GeneralFunction as rGeneralFunction
from ._rust import GeneralFunctionShapeInfo, GeneralFunctionSimpleSpec
from ._rust import Index as rIndex
from ._rust import Rearrange as rRearrange
from ._rust import Scalar as rScalar
from ._rust import Scatter as rScatter
from ._rust import Schedule as rSchedule
from ._rust import StoredCumulantVar as rStoredCumulantVar
from ._rust import Tag as rTag
from ._rust import TorchDeviceDtypeOp, scatter_to_concat
from .general_function_spec_base import GeneralFunctionSpecBase

MYPY = False


class CantConvertCircuitError(Exception):
    def __init__(self):
        try:
            import hypothesis
        except:
            super().__init__()
            return
        if hypothesis.currently_in_test_context():
            hypothesis.assume(False)
        super().__init__()


class FromPyGeneralFunctionSpec(GeneralFunctionSpecBase):
    name_val: str
    function_val: Callable  # Callable[[*torch.Tensor], torch.Tensor]
    get_shape_val: Callable  # Callable[[*Shape], Optional[Sequence[int]]]
    num_non_batchable_output_dims: int
    input_batchability: Tuple[bool, ...]

    def __init__(
        self,
        name: str,
        function: Callable,
        get_shape: Callable,
        num_non_batchable_output_dims: int,
        input_batchability: Tuple[bool, ...],
    ) -> None:
        self.name_val = name
        if not MYPY:  # this works on newer versions of MYPY - likely a bug
            self.function_val = function
            self.get_shape_val = get_shape
        self.num_non_batchable_output_dims = num_non_batchable_output_dims
        self.input_batchability = input_batchability

    @property
    def name(self) -> str:
        return self.name_val

    def compute_hash_bytes(self) -> bytes:
        from interp.circuit.eq_by_big_hash import hash_add, hash_add_by_id

        m = hashlib.blake2b(str(self.__class__).encode())
        hash_add(m, self.name)
        hash_add_by_id(m, self.function_val)
        hash_add_by_id(m, self.get_shape_val)
        hash_add(m, self.num_non_batchable_output_dims)
        hash_add(m, self.input_batchability)

        return m.digest()

    def function(self, *tensors: torch.Tensor) -> torch.Tensor:
        return self.function_val(*tensors)

    def get_shape_info(self, *shapes: "Shape") -> GeneralFunctionShapeInfo:
        return GeneralFunctionShapeInfo(
            self.get_shape_val(*shapes), self.num_non_batchable_output_dims, self.input_batchability
        )


def py_to_rust(py: "Circuit", device_dtype_op=TorchDeviceDtypeOp()) -> rCircuit:
    from interp.circuit.circuit import Circuit
    from interp.circuit.circuit_compiler.util import TorchAxisIndex
    from interp.circuit.circuit_utils import device_eq
    from interp.circuit.computational_node import (
        Add,
        Concat,
        Einsum,
        GeneralFunction,
        Index,
        UnaryRearrange,
        WildFunction,
    )
    from interp.circuit.constant import ArrayConstant, FloatConstant, One, Zero
    from interp.circuit.cumulant import Cumulant
    from interp.circuit.var import AutoTag, DiscreteVar, StoredCumulantVar

    @functools.cache
    def recurse(py: Circuit) -> rCircuit:
        if isinstance(py, (Zero, One, FloatConstant)):
            return rScalar(py.value, py.shape, py.name)
        elif isinstance(py, ArrayConstant):
            tensor = py.value
            if tensor.shape != py.shape:
                tensor = torch.broadcast_to(py.value, py.shape)
            if (
                device_dtype_op.dtype is not None
                and (torch_dtype := getattr(torch, device_dtype_op.dtype)) != tensor.dtype
            ):
                tensor = tensor.to(dtype=torch_dtype)
            if not device_eq(tensor, device_dtype_op.device):
                tensor = tensor.to(device=device_dtype_op.device)
            return rArray(tensor, py.name)
        elif isinstance(py, Add):
            return rAdd(*[recurse(x) for x in py.to_unweighted().items.keys()], name=py.name)
        elif isinstance(py, Einsum):
            return rEinsum(*[(recurse(operand), ints) for operand, ints in py.args], out_axes=py.out_axes, name=py.name)
        elif isinstance(py, UnaryRearrange):
            spec = py.get_spec()
            rust_child = recurse(py.node)
            return rRearrange(rust_child, spec.to_rust(), py.name)
        elif isinstance(py, Index):
            index = py.index
            new_idx: List[TorchAxisIndex] = []
            # convert slices from python approach
            for i, s in zip(index, py.node.shape):
                if (not MYPY) and isinstance(i, slice):
                    fix = lambda item: op.map(item, lambda x: max(min(x, s), -s))
                    i = slice(fix(i.start), fix(i.stop))
                elif (not MYPY) and isinstance(i, torch.Tensor) and i.ndim == 0:
                    i = int(i.item())

                new_idx.append(i)
            index = tuple(new_idx)

            if device_dtype_op.device is not None:
                index = tuple(
                    [
                        x.to(device=device_dtype_op.device)
                        if isinstance(x, torch.Tensor) and not device_eq(x, device_dtype_op.device)
                        else x
                        for x in index
                    ]
                )
            return rIndex(recurse(py.node), index, py.name)
        elif isinstance(py, Concat):
            return rConcat(*[recurse(x) for x in py.circuits], axis=py.axis, name=py.name)
        elif isinstance(py, AutoTag):
            return rTag(recurse(py.node), py.uuid, py.name)
        elif isinstance(py, GeneralFunction):
            recursed = recurse(py.node)
            if py.rust_spec is not None:
                return rGeneralFunction(recursed, spec=py.rust_spec, name=py.name)
            name = py.function.__name__.removesuffix("_fn") if hasattr(py.function, "__name__") else "unk_py_fn"
            result = rGeneralFunction.new_by_name_op(recursed, spec_name=name, name=py.name)
            if result is not None:
                return result
            else:
                num_non_batchable = len(py.normalized_non_batch_dims())

                gspec = FromPyGeneralFunctionSpec(
                    name,
                    py.function,
                    py_generalfunction_get_shape(num_non_batchable),
                    num_non_batchable,
                    (py.allows_batching,),
                )
                return rGeneralFunction(recursed, spec=gspec, name=py.name)
        elif isinstance(py, WildFunction):
            # TODO: not well tested!!!
            rec_nodes = [recurse(n) for n in py.nodes]

            gspec = FromPyGeneralFunctionSpec(
                py.function.get_name()
                if hasattr(py.function, "get_name")
                else py.function.__name__.removesuffix("_fn"),
                py.function,
                py.get_wild_function_shape_getter(),
                py.num_non_batchable_output_dims,
                tuple(py.input_batchability),
            )
            return rGeneralFunction(*rec_nodes, spec=gspec, name=py.name)
        elif isinstance(py, Cumulant):
            return rCumulant(*[recurse(x) for x in py.circuits], name=py.name)
        elif isinstance(py, DiscreteVar):
            return rDiscreteVar(recurse(py.values), recurse(py.probs_and_group), py.name)
        elif isinstance(py, StoredCumulantVar):
            return rStoredCumulantVar.new_mv(
                recurse(py.mean),
                recurse(py.cov),
                {k: recurse(v) for k, v in py.higher_cumulants.items()},
                py.uuid,
                py.name,
            )
        else:
            raise NotImplementedError(f"py_to_rust for {py.__class__.__name__} unimplemented")

    return recurse(py)


def rust_to_py(rust: rCircuit):
    from interp.circuit.circuit import Circuit
    from interp.circuit.computational_node import Add, Concat, Einsum, GeneralFunction, Index, UnaryRearrange
    from interp.circuit.constant import ArrayConstant, FloatConstant
    from interp.circuit.cumulant import Cumulant
    from interp.circuit.var import AutoTag, DiscreteVar, StoredCumulantVar

    @functools.cache
    def recurse(rust: rCircuit):
        result = recurse_raw(rust)
        assert result.shape == rust.shape, (result.shape, rust.shape, rust, result)
        assert result.ndim == len(rust.shape)
        return result

    def recurse_raw(rust: rCircuit) -> Circuit:
        if isinstance(rust, rScalar):
            return FloatConstant(rust.value, rust.shape, rust.name)
        elif isinstance(rust, rArray):
            return ArrayConstant(rust.value, rust.value.shape, name=rust.name)
        elif isinstance(rust, rAdd):
            return Add.from_unweighted_list([recurse(x) for x in rust.children], rust.name)
        elif isinstance(rust, rEinsum):
            return Einsum.from_axes_tuples(
                *[(recurse(operand), ints) for operand, ints in rust.args], out_axes=rust.out_axes, name=rust.name
            )
        elif isinstance(rust, rRearrange):
            return UnaryRearrange.from_spec(
                recurse(rust.node),
                rust.spec.to_py_rearrange_spec(rust.node.shape),
                rust.name,
            )
        elif isinstance(rust, rIndex):
            return Index(recurse(rust.node), tuple(rust.idx), name=rust.name)
        elif isinstance(rust, rGeneralFunction):
            spec = rust.spec
            if not (
                isinstance(spec, (GeneralFunctionSimpleSpec, FromPyGeneralFunctionSpec))
                and (rust.children[0].shape == rust.shape)
                and (len(rust.children) == 1)
            ):
                raise CantConvertCircuitError()

            if isinstance(spec, GeneralFunctionSimpleSpec):
                allows_batching = True
                function = spec.get_function()
            elif isinstance(spec, FromPyGeneralFunctionSpec):
                allows_batching = spec.input_batchability[0]
                function = spec.function_val
            else:
                assert_never(spec)

            return GeneralFunction(
                recurse(rust.children[0]),
                function,
                None,
                name=rust.name,
                allows_batching=allows_batching,  # this is always batchable
                non_batch_dims=tuple(range(-spec.num_non_batchable_output_dims, 0)),
                rust_spec=spec,
            )
        elif isinstance(rust, rConcat):
            return Concat(tuple([recurse(x) for x in rust.children]), rust.axis, rust.name)
        elif isinstance(rust, rScatter):
            return recurse(scatter_to_concat(rust))
        elif isinstance(rust, rTag):
            return AutoTag(recurse(rust.node), rust.uuid, rust.name)
        elif isinstance(rust, rDiscreteVar):
            return DiscreteVar(recurse(rust.values), recurse(rust.probs_and_group), rust.name)
        elif isinstance(rust, rStoredCumulantVar):
            highers = {k: recurse(v) for k, v in rust.cumulants.items() if k != 1 and k != 2}
            return StoredCumulantVar(
                recurse(rust.cumulants[1]), recurse(rust.cumulants[2]), highers, rust.uuid, rust.name
            )
        elif isinstance(rust, rCumulant):
            return Cumulant(tuple([recurse(x) for x in rust.children]), rust.name)
        elif isinstance(rust, rc.Module):
            raise CantConvertCircuitError()
        else:
            raise NotImplementedError(rust)

    return recurse(rust)


def py_generalfunction_get_shape(num_non_batchable: int):
    def check(shape: "Shape"):
        assert len(shape) >= num_non_batchable
        return shape

    return check


# this takes python circuit, rust version takes rust circuit
def schedule_replace_circuits(schedule: rSchedule, map: Dict["Circuit", torch.Tensor]):
    new_dict = {py_to_rust(k).hash: v for k, v in map.items()}
    result = schedule.replace_tensors(new_dict)
    return result


def rust_get_f64_evaluator():
    from interp.circuit.circuit import MemoizedFn
    from interp.circuit.circuit_utils import evaluate_fn

    return lambda circuits: [MemoizedFn(evaluate_fn(dtype=torch.float64))(c) for c in circuits]


def evaluate_py(pycirc):
    return py_to_rust(pycirc).evaluate()


def eval_opt_f64_py(pycirc):
    return rc.optimize_and_evaluate(
        py_to_rust(pycirc, TorchDeviceDtypeOp(dtype="float64")), rc.OptimizationSettings()
    ).to(dtype=torch.float64)


def cached_circuit_by_hash(fn):
    """Caches a function Circuit->Any, only storing hash of input"""
    cachy: Dict[Tuple, Any] = {}

    @functools.wraps(fn)
    def wrapped_fn(circuit: rCircuit, *other_args):
        nonlocal cachy
        key = tuple([circuit.hash, *other_args])
        if key in cachy:
            return cachy[key]
        result = fn(circuit, *other_args)
        cachy[key] = result
        return result

    return wrapped_fn


def cached_circuit_by_hash_weak(fn):
    """Caches a function Circuit->Any, only storing hash of input"""
    cachy: Dict[Tuple, weakref.ref[Any]] = {}

    @functools.wraps(fn)
    def wrapped_fn(circuit: rCircuit, *other_args):
        nonlocal cachy
        key = tuple([circuit.hash, *other_args])
        if key in cachy and (cached_result := cachy[key]()) is not None:
            return cached_result
        result = fn(circuit, *other_args)
        cachy[key] = weakref.ref(result)
        return result

    return wrapped_fn



from __future__ import annotations

import functools
import json
import random
from copy import copy
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, TypeVar, Union, cast
from uuid import UUID, uuid4

import cattrs
import torch
from attrs import frozen
from cattrs.preconf.json import make_converter as make_json_converter

import rust_circuit.optional as op

from ._rust import (
    Add,
    Array,
    Circuit,
    Einsum,
    GeneralFunction,
    Matcher,
    MatcherIn,
    Module,
    ModuleArgSpec,
    ModuleSpec,
    Parser,
    PrintOptions,
    Scalar,
    Symbol,
    TorchDeviceDtypeOp,
    deep_module_remove_unused_inputs,
    get_free_symbols,
    module_new_bind,
    simp,
)
from .py_utils import I

P = Parser()


# WARNING: Not tested for correctness!

# naming convention: module 'spec_circuit' has name 'x' while children are 'x.y'
# weights are 'x.w.y' ('w' isn't a placeholder). Constants are 'x.c.y'.
# then, callers can do 'z.x' or 'x.call' (or similar)


def to_name_dict(*circs: Circuit):
    out = {c.name: c for c in circs}
    assert len(out) == len(circs)
    return out


def add_new_circs(s: str, circs: Dict[str, Circuit], parser: Optional[Parser] = None):
    parser = op.unwrap_or(parser, Parser())
    parser.reference_circuits = {**parser.reference_circuits, **circs}

    return {**circs, **to_name_dict(*parser.parse_circuits(s))}


def make_spec(
    body: Circuit, order: List[Circuit] = [], exclude: Union[Callable[[Symbol], bool], MatcherIn] = lambda _: False
):
    exclude_matcher = Matcher(cast(MatcherIn, exclude))
    return ModuleSpec(
        body,
        [
            ModuleArgSpec(x)
            for x in sorted(get_free_symbols(body), key=lambda c: order.index(c) if c in order else -1)
            if not exclude_matcher(x)
        ],
    )


@frozen
class ModuleLibraryRet:
    circs: Dict[str, Circuit]
    body: Circuit
    spec: ModuleSpec
    name: str

    @classmethod
    def mk(cls, circs: Dict[str, Circuit], name: str) -> ModuleLibraryRet:
        body = circs[name]
        spec = make_spec(body, order=list(circs.values()))
        return cls(circs=circs, body=body, spec=spec, name=name)


def zero_input(circs: Dict[str, Circuit], to_zero: str):
    return to_name_dict(
        *(
            # we could make this faster if we needed
            deep_module_remove_unused_inputs(simp(c.update(to_zero, lambda _: Scalar(0.0, shape=())))).update(
                True, lambda x: x.rename(x.name.removesuffix(" rem_unused"))
            )
            for s, c in circs.items()
            if s != to_zero
        )
    )


layernorm_str = f"""
'ln.y' Einsum h,->h
  'ln.mean_subbed' Add
    'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
    'ln.neg_mean' Einsum h,z,->z # z is padding dim for add
      'ln.input'
      'ln.neg' [1] Scalar -1
      'ln.c.recip_hidden_size' GeneralFunction reciprocal
        'ln.c.hidden_size' GeneralFunction last_dim_size
          'ln.input'
  'ln.rsqrt' GeneralFunction rsqrt
    'ln.var_p_eps' Add
      'ln.c.eps' [] Scalar 1e-5
      'ln.var' Einsum h,h,->
        'ln.mean_subbed'
        'ln.mean_subbed'
        'ln.c.recip_hidden_size'

'ln.input'

'ln' Add
  'ln.w.bias' [?] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487
  'ln.y_scaled' Einsum h,h->h
    'ln.y'
    'ln.w.scale' [?] Symbol 0fa341c3-34b3-4699-847f-08674808b28a

'ln.w.bias'
'ln.w.scale'
"""

layernorm_circs = add_new_circs(layernorm_str, {})
layernorm = ModuleLibraryRet.mk(layernorm_circs, "ln")

batchnorm_str = f"""
'bn.y' Einsum h,h->h
  'bn.mean_subbed' Add
    'bn.input' [?] Symbol cfe0aa25-1214-4d3e-ad53-417fedbb9b7b
    'bn.neg_mean' Einsum h,->h
      'bn.mean' [?] Symbol 217403c9-03b6-4444-8f72-de03b95229d0
      'bn.neg' [] Scalar -1
  'bn.rsqrt' GeneralFunction rsqrt
    'bn.var_p_eps' Add
      'bn.c.eps' [] Scalar 1e-5
      'bn.var' [?] Symbol c4e41ca7-e3ab-429e-81b6-25bb14423fb4

'bn.input'
'bn.mean'
'bn.var'

# same as ln above, we could dedudup if we wanted
'bn' Add
  'bn.w.bias' [?] Symbol 3e861c08-9a74-4348-87df-3752a8557aea
  'bn.y_scaled' Einsum h,h->h
    'bn.y'
    'bn.w.scale' [?] Symbol ccbac7e0-c931-422b-9724-1e24ab6e9c63

'bn.w.bias'
'bn.w.scale'
"""

batchnorm_circs = add_new_circs(batchnorm_str, {})
batchnorm = ModuleLibraryRet.mk(batchnorm_circs, "bn")

not_mask_str = """
'not_mask' Add
  'one' [] Scalar 1.
  'not_mask.neg_mask' Einsum ,->
    'not_mask.input' [] Symbol b46f6370-11e1-4535-aabc-94554c234673
    'neg_one' [] Scalar -1.

'not_mask.input'
"""

not_mask_circs = add_new_circs(not_mask_str, {})
not_mask = ModuleLibraryRet.mk(not_mask_circs, "not_mask")

raw_attention_str = f"""
'a.head' Einsum sV,dV->sd
  'a.comb_v' Einsum qk,kV->qV
    'a.attn_probs' GeneralFunction softmax
      'a.attn_scores' Add
        'a.attn_scores_raw' [?, ?] Einsum qc,kc,,qk->qk
          'a.q_p_bias' [?, ?] Add
            'a.q' Einsum qd,cd->qc
              'a.q.input' [?, ?] Symbol 4f80d1a1-86a4-4e44-94f7-909ec7089061
              'a.w.q_h' [?, ?] Symbol 665efa60-d86c-40d5-92b2-b96d11686a8b
            'a.w.q_bias_h' [?] Symbol 7d531f53-6cce-4bf3-82db-5fe8b2eef974
          'a.k_p_bias' [?, ?] Add
            'a.k' Einsum kd,cd->kc
              'a.k.input' [?, ?] Symbol 664bddee-28ca-47e7-9fb7-9a718de06619
              'a.w.k_h' [?, ?] Symbol 41177709-446d-4588-b9e5-c2bbf59d53a0
            'a.w.k_bias_h' [?] Symbol a891aae4-3c24-4713-afc7-8b954c6fc1b5
          'a.c.div_head_size' GeneralFunction rsqrt
            'a.c.head_size' GeneralFunction last_dim_size
              'a.c.bias_for_head_size' [?] Einsum jk->j # size on this line is just asserted
                'a.w.k_h' [?, ?]
          # mask is true at where positions *are* allowed. (This differs from old pos_mask code)
          'a.mask' [?, ?] Symbol ccfe5bc9-b402-42dd-a5e1-191e6fb7c268
        'a.score_neg_inf_bias' Einsum qk,->qk
          'a.not_mask' Module
            'not_mask'
            'a.mask' ! 'not_mask.input'
          'a.neg_inf' [] Scalar -10_000.0
    'a.v_p_bias' Add
      'a.v' Einsum kd,Vd->kV
        'a.v.input' [?, ?] Symbol 8fd4c632-7f28-49ee-84cc-3dde997e0693
        'a.w.v_h' [?, ?] Symbol 79b6ebff-f9d0-411a-bcdc-530cc13e1524
      'a.w.v_bias_h' [?] Symbol dfb2c4ec-9378-40ee-a360-7d58f2b96954
  'a.w.o_h' [?, ?] Symbol 11a116cb-2168-4725-a06f-1b61a8ca6797

'a.q.input'
'a.k.input'
'a.v.input'
'a.mask'
'a.w.q_h'
'a.w.q_bias_h'
'a.w.k_h'
'a.w.k_bias_h'
'a.w.v_h'
'a.w.v_bias_h'
'a.w.o_h'


'a.head.on_inp' Module
  'a.head'
  'a.qk_input' Add ! 'a.q.input'
    'a.input' [?, ?] Symbol f9eabd07-e2ab-4ed4-8b4a-c9c039d61835
    'a.pos_input' [?, ?] Symbol eab8313e-d910-4174-8dbe-c612954eec34
  'a.qk_input' ! 'a.k.input'
  'a.input' ! 'a.v.input'

'a.input'
'a.pos_input'

'a.p_bias' Add
  'a' Einsum hsd->sd # reduce head dim
    # batch over head
    'a.heads' Module
      'a.head.on_inp'
      'a.w.q' [?, ?, ?] Symbol cf8f9c58-1875-45b0-8007-66b66b8a405a ! 'a.w.q_h'
      'a.w.q_bias' [?, ?] Symbol e896a43e-ba35-43d5-97e7-e189caf3278b ! 'a.w.q_bias_h'
      'a.w.k' [?, ?, ?] Symbol c342f1cd-71e6-4848-86b4-1c3ffdd46753 ! 'a.w.k_h'
      'a.w.k_bias' [?, ?] Symbol 13465b29-c28a-4d55-8a27-22c26cc01c69 ! 'a.w.k_bias_h'
      'a.w.v' [?, ?, ?] Symbol a90db69c-f4ad-47d6-8b76-faa4b107dacd ! 'a.w.v_h'
      'a.w.v_bias' [?, ?] Symbol ea64ee26-2438-4aa9-b777-0ef4f5f70e74 ! 'a.w.v_bias_h'
      'a.w.o' [?, ?, ?] Symbol 24e0e5cf-8b68-4bf2-b198-17237523b237 ! 'a.w.o_h'
  'a.w.o_bias' [?] Symbol 61d54061-63e2-4a5d-9c7a-15c1ecc53b95

'a.w.q'
'a.w.q_bias'
'a.w.k'
'a.w.k_bias'
'a.w.v'
'a.w.v_bias'
'a.w.o'
'a.w.o_bias'


# We split into two pieces so we can apply rotary positional embeddings to only the first.
# Also, we use HALF_ROTARY_DIM to enforce that the rotary dim is even.
'qk_with_rot' [?, ?] Rearrange q s:2 e:2 rh -> q (s:2 e:2 rh)
  'qk_with_rot.cat' [?, 2, 2, ?] Concat 1
    'qk_with_rot.q_rot' Rearrange q e rh -> q 1 e rh
      'qk_rot' [?, 2, ?] GeneralFunction apply_rotary_pos_emb at rust_circuit.generalfuncs.rotary_pos_emb:ApplyRotaryPosEmb
        'qk_rot.inp' Index [:, 0]
          'qk_split' [?, 2, 2, ?] Rearrange q (s:2 e:2 rh) -> q s:2 e:2 rh
            # q or k
            'qk' [?, ?] Symbol 442c7d04-1948-4904-b96a-5c0acf22f19c
    'qk_with_rot.q_pass' Index [:, 1:]
      'qk_split'

# Most of the below is copied from a.head; could factor into a module but backcompat is annoying

'a_rope.head' Einsum sV,dV->sd
  'a_rope.comb_v' Einsum qk,kV->qV
    'a_rope.attn_probs' GeneralFunction softmax
      'a_rope.attn_scores' Add
        'a_rope.attn_scores_raw' Einsum qc,kc,,qk->qk
          'a_rope.q' Module
            'qk_with_rot'
            'a.q_p_bias' ! 'qk'
          'a_rope.k' Module
            'qk_with_rot'
            'a.k_p_bias' ! 'qk'
          'a.c.div_head_size'
          'a.mask'
        'a.score_neg_inf_bias'
    'a.v_p_bias'
  'a.w.o_h'

'a_rope.head.on_inp' Module
  'a_rope.head'
  'a.qk_input' ! 'a.q.input'
  'a.qk_input' ! 'a.k.input'
  'a.input' ! 'a.v.input'

'a_rope.p_bias' Add
  'a_rope' Einsum hsd->sd # reduce head dim
    # batch over head
    'a_rope.heads' Module
      'a_rope.head.on_inp'
      'a.w.q' ! 'a.w.q_h'
      'a.w.q_bias' ! 'a.w.q_bias_h'
      'a.w.k' ! 'a.w.k_h'
      'a.w.k_bias' ! 'a.w.k_bias_h'
      'a.w.v' ! 'a.w.v_h'
      'a.w.v_bias' ! 'a.w.v_bias_h'
      'a.w.o' ! 'a.w.o_h'
  'a.w.o_bias'
"""
raw_attention_circs = add_new_circs(raw_attention_str, not_mask_circs)


def get_attention(bias: bool = False, pos: bool = False, rope: bool = False):
    circs = copy(raw_attention_circs)
    if not bias:
        for weight in ["a.w.q_bias", "a.w.k_bias", "a.w.v_bias", "a.w.o_bias"]:
            circs = zero_input(circs, weight + "_h")
            circs = zero_input(circs, weight)

    if not pos:
        circs = zero_input(circs, "a.pos_input")

    circ = "a"
    if rope:
        circ += "_rope"
    if bias:
        circ += ".p_bias"
    return ModuleLibraryRet.mk(circs, circ)


m_input = Symbol((None,), UUID("13ec4cdc-5f25-4969-870b-0bfa2300187b"), name="m.input")
m_base_circs: Dict[str, Circuit] = {"m.input": m_input}


raw_bilinear_mlp_str = f"""
'm.p_bias' Add
  'm' Einsum h,h,oh->o
    'm.pre0' Index [0,:]
      'm.fold_pre' Rearrange (a:2 b) -> a:2 b
        'm.pre_p_bias' Add
          'm.pre' Einsum i,hi->h
            'm.input'
            'm.w.proj_in' [?, ?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
          'm.w.in_bias' [?] Symbol 886b5425-cd19-4db4-871a-c46cb1a23114
    'm.pre1' Index [1,:]
      'm.fold_pre'
    'm.w.proj_out' [?, ?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
  'm.w.out_bias' [?] Symbol 7efddf2e-20af-492c-a94e-1d19468f333f

'm.w.proj_in'
'm.w.in_bias'
'm.w.proj_out'
'm.w.out_bias'
"""

raw_bilinear_mlp_circs = add_new_circs(raw_bilinear_mlp_str, m_base_circs)


@functools.cache
def get_bilinear_mlp(output_bias: bool = False):
    if output_bias:
        circs = raw_bilinear_mlp_circs
    else:
        circs = zero_input(raw_bilinear_mlp_circs, "m.w.out_bias")

    return ModuleLibraryRet.mk(circs, "m.p_bias" if output_bias else "m")


@functools.cache
def get_pointwise_mlp(function_str: str = "gelu", output_bias: bool = False):
    s = f"""
    'm.p_bias' Add
      'm' Einsum h,oh->o
        'm.act' GeneralFunction {function_str}
          'm.pre' Add
            'm.pre_mul' Einsum i,hi->h
              'm.input'
              'm.w.proj_in' [?, ?] Symbol 5217f963-0cdb-460e-bb1f-f82f7fbb3cd9
            'm.w.in_bias' [?] Symbol c870ec00-8c6f-4080-907c-703ea85dde48
        'm.w.proj_out' [?, ?] Symbol fdefa9af-a7d6-4a38-a7ed-5ce816c6efe7
      'm.w.out_bias' [?] Symbol 113c30ba-fa88-4f34-a301-ea7912e03064

    'm.input'
    'm.w.proj_in'
    'm.w.in_bias'
    'm.w.proj_out'
    'm.w.out_bias'
    """
    circs = add_new_circs(s, m_base_circs)
    if not output_bias:
        circs = zero_input(circs, "m.w.out_bias")

    return ModuleLibraryRet.mk(circs, "m.p_bias" if output_bias else "m")


def get_norm_bind(prefix: str, uuids: List[UUID], norm_type: Literal["ln", "bn"]):
    uuids = copy(uuids)
    p = prefix
    nt = norm_type
    bind_str = f"""
    '{p}.norm' Module
      '{nt}'
      '{p}.norm.input' [?, ?] Symbol {uuids.pop(4)} ! '{nt}.input'"""

    if norm_type == "bn":
        bind_str += f"""
      '{p}.{nt}.mean' [?] Symbol {uuids.pop(3)} ! '{nt}.mean'
      '{p}.{nt}.var' [?] Symbol {uuids.pop(2)} ! '{nt}.var'"""

    bind_str += f"""
      '{p}.{nt}.w.bias' [?] Symbol {uuids.pop(1)} ! '{nt}.w.bias'
      '{p}.{nt}.w.scale' [?] Symbol {uuids.pop(0)} ! '{nt}.w.scale'

    '{p}.norm.input'"""

    if norm_type == "bn":
        bind_str += f"""
    '{p}.{nt}.mean'
    '{p}.{nt}.var'"""

    bind_str += f"""
    '{p}.{nt}.w.bias'
    '{p}.{nt}.w.scale'
    """
    circs = add_new_circs(bind_str, layernorm_circs if norm_type == "ln" else batchnorm_circs)
    return ModuleLibraryRet.mk(circs, f"{p}.norm")


def print_uuid_list_repr(n: int):
    """utility for writing this code"""
    print(repr([uuid4() for _ in range(n)]))


m_norm_uuids = [
    UUID("fe330d51-0164-4b49-a504-81769d550bb1"),
    UUID("771e2d50-3414-459e-87ff-879f83d4a15c"),
    UUID("bf7ea689-c481-4ad4-89d0-0a59cd476739"),
    UUID("cf2adbd9-b04f-42af-86d7-cc5c8890607b"),
    UUID("f63f8838-9d69-4da2-a902-8eaf325f07a7"),
]
a_norm_uuids = [
    UUID("c564149d-e226-4e3d-8e47-7d6e2ceea99e"),
    UUID("2c737289-2702-404c-a22e-ad37c2652620"),
    UUID("33774cb3-f047-4a1e-ac0c-e6f8c93d2bb2"),
    UUID("981982a3-253b-4c0d-8789-c0bc9dcd229b"),
    UUID("6a622698-fd68-4d25-aeee-e8d38e68049e"),
]
final_norm_uuids = [
    UUID("2110aef7-70d5-4f77-a929-044e40c31cbd"),
    UUID("10652dc2-e9a8-4f1e-8bad-99d46804c7b3"),
    UUID("6cfcb4e6-7081-4d2d-bef9-1e9db52d4444"),
    UUID("94e8bc7f-1689-429f-bfec-67d33965aef5"),
    UUID("0851aebd-0d39-4f65-89f9-8a8afaed631a"),
]

m_ln_bind = get_norm_bind("m", m_norm_uuids, "ln")
a_ln_bind = get_norm_bind("a", a_norm_uuids, "ln")
p_ln_bind = get_norm_bind("p", a_norm_uuids, "ln")
m_bn_bind = get_norm_bind("m", m_norm_uuids, "bn")
a_bn_bind = get_norm_bind("a", a_norm_uuids, "bn")
p_bn_bind = get_norm_bind("p", a_norm_uuids, "bn")
final_ln_bind = get_norm_bind("final", final_norm_uuids, "ln")
final_bn_bind = get_norm_bind("final", final_norm_uuids, "bn")


# TODO: should this be a module?
def get_norm_call(prefix: str, circs: Dict[str, Circuit], mod: Optional[str] = None):
    norm_call_str = f"""
    '{prefix}.norm_call' Module
      '{op.unwrap(mod, prefix)}'
      '{prefix}.norm' ! '{prefix}.input'
    """

    circs = add_new_circs(norm_call_str, circs)
    return ModuleLibraryRet.mk(circs, f"{prefix}.norm_call")


b_input = Symbol((None, None), UUID("5837c4fd-f5ac-4bff-8456-abf3e95bcf36"), name="b.input")


def rename_circs_to_keys(d: Dict[str, Circuit], rename_suffix: str = ""):
    return {s: c.rename(s + rename_suffix) for s, c in d.items()}


T = TypeVar("T")


def apply_prefix(d: Dict[str, T], prefix: str) -> Dict[str, T]:
    return {f"{prefix}.{s}": c for s, c in d.items()}


def add_number(name: str, num: int):
    before, _, after = name.partition(".")
    return f"{before}{num}.{after}"


MlpActType = Literal["gelu", "gelu_new", "relu", "bilinear"]
SoftmaxType = Literal["softmax"]
PosEncType = Literal["gpt", "shortformer", "none"]


@frozen
class TransformerBlockParams:
    norm_type: Optional[Literal["ln", "bn"]] = "ln"
    attn_bias: bool = False
    attn_pos: bool = False
    use_mlp: bool = True
    mlp_act_type: MlpActType = "gelu"  # type: ignore
    mlp_output_bias: bool = False
    use_parallel_residual: bool = False  # currently implies only one shared norm
    use_rope: bool = False

    @functools.cache
    def get(self):
        # order for circs matters
        circs: Dict[str, Circuit] = {}
        if self.use_mlp:
            if self.mlp_act_type == "bilinear":
                mlp = get_bilinear_mlp(output_bias=self.mlp_output_bias)
                circs = {**circs, **mlp.circs}
            else:
                mlp = get_pointwise_mlp(function_str=self.mlp_act_type, output_bias=self.mlp_output_bias)
                circs = {**circs, **mlp.circs}
            if self.norm_type is not None:
                circs = {**circs, **(m_ln_bind if self.norm_type == "ln" else m_bn_bind).circs}
            if self.norm_type is None:
                mlp_with_norm = mlp
            else:
                mlp_with_norm = get_norm_call("m", circs, mod=mlp.name)
                circs = mlp_with_norm.circs
        if self.use_parallel_residual:
            circs = {**circs, **(p_ln_bind if self.norm_type == "ln" else p_bn_bind).circs}

        attn = get_attention(bias=self.attn_bias, pos=self.attn_pos, rope=self.use_rope)
        circs = {**circs, **attn.circs}
        if self.norm_type is not None:
            circs = {**circs, **(a_ln_bind if self.norm_type == "ln" else a_bn_bind).circs}
        if self.norm_type is None:
            attn_with_norm = attn
        else:
            attn_with_norm = get_norm_call("a", circs, mod=attn.name)
            circs = attn_with_norm.circs

        circs = {**circs, "b.input": b_input}

        attn_only_str = f"""
        'b.resid.a' Add
          'b.a' Module
            '{attn_with_norm.name}'
            'b.input' ! 'a{'.norm' if self.norm_type is not None else ''}.input'
          'b.input'

        'b.input'
        """

        cur_resid_name = "b.resid.a"

        circs = add_new_circs(attn_only_str, circs)

        if self.use_parallel_residual:
            assert self.use_mlp and self.norm_type is not None
            block_str = f"""
          'b.add' Add
            'b.a' Module
              '{attn.name}'
              'p.input' [?, ?] Symbol e76c1fd7-a230-4295-9199-17dc8dad79d0 ! 'a.input'
            'b.m' Module
              '{mlp.name}'
              'p.input' ! 'm.input'

          'p.input'
          """
            circs = add_new_circs(block_str, circs)
            circs = get_norm_call("p", circs, mod="b.add").circs
            block_str = """
          'b.resid.p' Add
            'b.p' Module
              'p.norm_call'
              'b.input' ! 'p.norm.input'
            'b.input'
          """
            circs = add_new_circs(block_str, circs)
            cur_resid_name = "b.resid.p"
        elif self.use_mlp:
            mlp_input = cur_resid_name
            block_str = f"""
            'b.resid.m' Add
              'b.m' Module
                '{mlp_with_norm.name}'
                '{mlp_input}' ! 'm{'.norm' if self.norm_type is not None else ''}.input'
              '{cur_resid_name}'
            """

            cur_resid_name = "b.resid.m"

            circs = add_new_circs(block_str, circs)

        circs["b"] = circs[cur_resid_name].rename("b")

        return ModuleLibraryRet.mk(circs, "b")

    # add function for rebinding inputs also as needed
    @functools.cache
    def get_rebound_weighty(self, num: int):
        gotten = self.get()

        rd = random.Random()
        rd.seed(num)

        non_weighty_inputs = {"b.input", "a.pos_input", "a.mask"}

        circs = copy(gotten.circs)
        spec = make_spec(gotten.body, order=list(gotten.circs.values()), exclude=non_weighty_inputs)
        bound = Module.new_flat(
            spec,
            *[
                Symbol(
                    arg_spec.symbol.shape,
                    UUID(int=rd.getrandbits(128)),
                    add_number(arg_spec.symbol.name, num),
                )
                for arg_spec in spec.arg_specs
            ],
            name=f"b{num}",
        )

        assert bound.name not in circs
        circs[bound.name] = bound

        return ModuleLibraryRet.mk(circs, bound.name)

    def garbage_init_norm_weights(self, hidden_size: int, device_dtype: TorchDeviceDtypeOp = TorchDeviceDtypeOp()):
        """be warned this is not random init, it's GARBAGE, it won't train or behave properly at all"""

        weights: Dict[str, Circuit] = {}
        if self.norm_type is not None:
            if self.norm_type == "bn":
                weights.update(
                    {
                        f"{self.norm_type}.mean": Array.randn(hidden_size, device_dtype=device_dtype),
                        f"{self.norm_type}.var": Array(torch.randn(hidden_size).to(device=device_dtype.device) ** 2),
                    }
                )
            weights.update(
                {
                    f"{self.norm_type}.w.bias": Array.randn(hidden_size, device_dtype=device_dtype),
                    f"{self.norm_type}.w.scale": Array.randn(hidden_size, device_dtype=device_dtype),
                }
            )

        return rename_circs_to_keys(weights, rename_suffix="_arr")

    def garbage_init_weights(
        self, hidden_size: int, head_size: int, num_heads: int, device_dtype: TorchDeviceDtypeOp = TorchDeviceDtypeOp()
    ):
        """be warned this is not random init, it's GARBAGE, it won't train or behave properly at all"""

        weights: Dict[str, Circuit] = {}
        weights.update(
            apply_prefix(self.garbage_init_norm_weights(hidden_size=hidden_size, device_dtype=device_dtype), "a")
        )

        # if self.block_spec.attn_pos:
        #     weights.update({"a.pos_input": Array.randn(*batch_shape, seq_len, hidden_size)})

        weights.update(
            {
                # "a.mask": Array.randn(seq_len, seq_len, device_dtype=device_dtype),
                "a.w.q": Array.randn(num_heads, head_size, hidden_size, device_dtype=device_dtype),
                "a.w.k": Array.randn(num_heads, head_size, hidden_size, device_dtype=device_dtype),
                "a.w.v": Array.randn(num_heads, head_size, hidden_size, device_dtype=device_dtype),
                "a.w.o": Array.randn(num_heads, hidden_size, head_size, device_dtype=device_dtype),
            }
        )

        if self.attn_bias:
            weights.update(
                {
                    "a.w.q_bias": Array.randn(num_heads, head_size, device_dtype=device_dtype),
                    "a.w.k_bias": Array.randn(num_heads, head_size, device_dtype=device_dtype),
                    "a.w.v_bias": Array.randn(num_heads, head_size, device_dtype=device_dtype),
                    "a.w.o_bias": Array.randn(hidden_size, device_dtype=device_dtype),
                }
            )

        if self.use_mlp:
            weights.update(
                apply_prefix(self.garbage_init_norm_weights(hidden_size=hidden_size, device_dtype=device_dtype), "m")
            )

            weights.update(
                {
                    "m.w.proj_in": Array.randn(hidden_size * 4, hidden_size, device_dtype=device_dtype),
                    "m.w.in_bias": Array.randn(hidden_size * 4, device_dtype=device_dtype),
                    "m.w.proj_out": Array.randn(
                        hidden_size,
                        hidden_size * (2 if self.mlp_act_type == "bilinear" else 4),
                        device_dtype=device_dtype,
                    ),
                }
            )

            if self.mlp_output_bias:
                weights.update({"m.w.out_bias": Array.randn(hidden_size, device_dtype=device_dtype)})

        return rename_circs_to_keys(weights, "_arr")


# aka cross entropy, aka log loss
log_likelyhood_str = f"""
'll' GeneralFunction gen_index_at_0_batch_x_c
  'log_probs' GeneralFunction log_softmax
    'll.input' [?] Symbol b9d111e5-b793-4f63-84a0-2f8590b9f39c
  'll.label' [] Symbol 74c503b9-fe2c-4f8b-9350-7351a292c351

'll.input'
'll.label'

'nll' Einsum ,->
  'll'
  'nll.neg' [] Scalar -1
"""

log_likelyhood_circs = add_new_circs(log_likelyhood_str, {})
log_likelyhood = ModuleLibraryRet.mk(log_likelyhood_circs, "ll")
negative_log_likelyhood = ModuleLibraryRet.mk(log_likelyhood_circs, "nll")


# TODO: maybe add more bindings on top of this
@frozen
class TransformerParams:
    block_params: TransformerBlockParams = TransformerBlockParams()
    num_layers: int = 2
    use_norm_output: bool = True
    output_bias: bool = False

    @property
    def norm_type(self):
        return self.block_params.norm_type

    @functools.cache
    def get(self):
        blocks = [self.block_params.get_rebound_weighty(i) for i in range(self.num_layers)]
        # keeps right ordering
        circs = dict(x for b in blocks for x in b.circs.items())

        out: Circuit = Symbol((None, None), UUID("ece2bb5d-c6d6-4b7a-93a5-32b5ac264888"), "t.input")
        assert "t.input" not in circs
        circs["t.input"] = out
        for i, block in enumerate(blocks):
            out = module_new_bind(block.body, ("b.input", out), name=f"b{i}.call")
            circs[out.name] = out

        if self.norm_type is not None and self.use_norm_output:
            circs = {**circs, **(final_ln_bind if self.norm_type == "ln" else final_bn_bind).circs}
            final_norm_str = f"""
            'final.call' Module
              'final.norm'
              '{out.name}' ! 'final.norm.input'
            """
            circs = add_new_circs(final_norm_str, circs)
            out = circs["final.call"]

        logits_str = f"""
        't.logits' Einsum sh,vh->sv
          '{out.name}'
          't.w.unembed' [?, ?] Symbol 85d5a05a-ef9e-4910-a967-3f27951f67cf
        't.logits_p_bias' Add
          't.logits'
          't.w.unembed_bias' [?] Symbol 62d9f91a-5df2-4eb8-af75-5bdfb15eb974

        't.w.unembed'
        't.w.unembed_bias'
        """
        circs = add_new_circs(logits_str, circs)
        if not self.output_bias:
            circs = {s: c for s, c in circs.items() if s not in ["t.w.unembed_bias", "t.logits_p_bias"]}

        return ModuleLibraryRet.mk(circs, name="t.logits" + ("_p_bias" if self.output_bias else ""))

    def garbage_init_weights(
        self,
        hidden_size: int = 3,
        head_size: int = 5,
        num_heads: int = 7,
        vocab_size: int = 11,
        device_dtype: TorchDeviceDtypeOp = TorchDeviceDtypeOp(),
    ):
        """be warned this is not random init, it's GARBAGE, it won't train or behave properly at all"""
        weights: Dict[str, Circuit] = {}
        for i in range(self.num_layers):
            weights.update(
                {
                    add_number(s, i): c
                    for s, c in self.block_params.garbage_init_weights(
                        hidden_size=hidden_size, head_size=head_size, num_heads=num_heads, device_dtype=device_dtype
                    ).items()
                }
            )
        if self.use_norm_output:
            weights.update(
                apply_prefix(
                    self.block_params.garbage_init_norm_weights(hidden_size=hidden_size, device_dtype=device_dtype),
                    "final",
                )
            )

        weights["t.w.unembed"] = Array.randn(vocab_size, hidden_size, device_dtype=device_dtype)

        if self.output_bias:
            weights["t.w.unembed_bias"] = Array.randn(vocab_size, device_dtype=device_dtype)

        return rename_circs_to_keys(weights, rename_suffix="_arr")

    def garbage_call(
        self,
        hidden_size: int = 3,
        head_size: int = 5,
        num_heads: int = 7,
        seq_len: int = 9,
        vocab_size: int = 11,
        batch_shape: Tuple[int, ...] = (13, 2),
        device_dtype: TorchDeviceDtypeOp = TorchDeviceDtypeOp(),
    ):
        """be warned this is not random init, it's GARBAGE, it won't train or behave properly at all"""
        weights = self.garbage_init_weights(
            hidden_size=hidden_size,
            head_size=head_size,
            num_heads=num_heads,
            vocab_size=vocab_size,
            device_dtype=device_dtype,
        )

        # this could be made faster if we needed
        # seems fine for now
        bound_weights = module_new_bind(self.get().body, *weights.items(), name="t.bind_w")
        inputs = rename_circs_to_keys(
            {
                "t.input": Array.randn(*batch_shape, seq_len, hidden_size),
                "a.mask": Array.randn(seq_len, seq_len, device_dtype=device_dtype),
                **(
                    {"a.pos_input": Array.randn(*batch_shape, seq_len, hidden_size)}
                    if self.block_params.attn_pos
                    else {}
                ),
            },
            rename_suffix="_rand_inp",
        )
        return module_new_bind(bound_weights, *inputs.items(), name="t.call"), weights, inputs


@frozen
class TransformerInfo:
    params: TransformerParams
    model_class: str = "GPTBeginEndToks"
    # maybe below 2 shouldn't be optional?
    pos_enc_type: Optional[PosEncType] = "gpt"  # type: ignore
    causal_mask: Optional[bool] = True
    extra: Optional[Any] = None

    def dump_model_string(self, *circs: Circuit):
        info_json = make_json_converter().dumps(self)
        assert "\n" not in info_json
        prefix_str = f"# info:{info_json}\n"
        rep = PrintOptions(reference_circuits={c: s for s, c in self.params.get().circs.items()}).repr(
            *circs,
        )

        return prefix_str + rep + "\n"

    def bind_to_input(
        self,
        c: Circuit,
        inp_tok_embeds: Circuit,
        pos_embed_weights: Optional[Circuit] = None,
        inp_mask: Optional[Circuit] = None,
        inp_mask_has_q: bool = False,
        prefix: str = "t",
    ):
        seq_len: int = cast(int, inp_tok_embeds.shape[-2])

        if pos_embed_weights is None:
            pos_embeds = None
        else:
            pos_embeds = pos_embed_weights.index(
                I[
                    None:seq_len,
                ],
                name="t.w.pos_embeds_idxed",
            )
        inp: Circuit
        if self.pos_enc_type == "gpt":
            assert pos_embeds is not None
            inp = Add(inp_tok_embeds, pos_embeds, name=f"{prefix}.inp_tok_pos")
        else:
            inp = inp_tok_embeds

        mask: Circuit
        if self.causal_mask:
            mask = Array(
                (torch.arange(seq_len)[:, None] >= torch.arange(seq_len)[None, :]).to(
                    device=inp_tok_embeds.device, dtype=inp_tok_embeds.torch_dtype or torch.float32
                ),
                "t.a.c.causal_mask",  # prefix doesn't apply to constants
            )
        else:
            mask = Scalar(1.0, shape=(seq_len, seq_len), name="t.a.c.full_mask")

        if inp_mask is not None:

            class DimNumMaker:
                num: int = 0

                def __call__(self, n: int):
                    out = tuple(range(self.num, self.num + n))
                    self.num += n
                    return out

            m = DimNumMaker()
            [k] = m(1)
            [q] = m(1)
            batch = m(inp_mask.ndim - 1 - (1 if inp_mask_has_q else 0))
            # TODO: simplify out 1 in this case!
            mask = Einsum(
                (inp_mask, batch + ((q, k) if inp_mask_has_q else (k,))),
                (mask, (q, k)),
                out_axes=batch + (q, k),
                name=f"{prefix}.a.mask",
            )

        pairs = [("t.input", inp), ("a.mask", mask)]
        if self.pos_enc_type == "shortformer":
            assert pos_embeds is not None
            pairs.append(("a.pos_input", pos_embeds))

        return module_new_bind(c, *pairs, name=f"{prefix}.call")

    def bind_to_input_tokens(
        self,
        c: Circuit,
        toks: Circuit,
        tok_embed_weights: Circuit,
        pos_embed_weights: Circuit,
        inp_mask: Optional[Circuit] = None,
        inp_mask_has_q: bool = False,
        prefix: str = "t",
    ):
        embedded = GeneralFunction.gen_index(tok_embed_weights, toks, -2)
        return self.bind_to_input(c, embedded, pos_embed_weights, inp_mask, inp_mask_has_q, prefix)

    def bind_to_input_tokens_int16(
        self,
        c: Circuit,
        toks: torch.Tensor,
        tok_embed_weights: Circuit,
        pos_embed_weights: Circuit,
        inp_mask: Optional[Circuit] = None,
        inp_mask_has_q: bool = False,
        prefix: str = "t",
    ):
        tok_array = Array(toks)
        upcasted = Module(token_upcast_module, "toks_upcasted", **{"upcast_toks.int16_toks": tok_array})
        return self.bind_to_input_tokens(
            c, upcasted, tok_embed_weights, pos_embed_weights, inp_mask, inp_mask_has_q, prefix
        )


def load_transformer_model_string(
    s: str, parser: Optional[Parser] = None
) -> tuple[dict[str, Circuit], Any, TransformerInfo]:
    s = s.strip("\n")
    assert s.startswith("# info:")
    info_s, _, _ = s.partition("\n")

    info = cattrs.structure(json.loads(info_s.removeprefix("# info:")), TransformerInfo)

    circs = info.params.get().circs
    added = add_new_circs(s, circs, parser=parser)

    # copied out of model_loading to work in open source version
    try:
        from interp.model.model_loading import MODEL_CLASS_STR_TO_MODEL_AND_TOKENIZER_FNS

        tokenizer = MODEL_CLASS_STR_TO_MODEL_AND_TOKENIZER_FNS[info.model_class][1]()
    except ImportError:
        if info.model_class in {"GPT", "GPTBeginEndToks"}:
            import transformers

            tokenizer = transformers.GPT2TokenizerFast.from_pretrained("gpt2")
            tokenizer._add_tokens(["[END]"])
            tokenizer.pad_token = "[END]"
            if info.model_class == "GPTBeginEndToks":
                tokenizer._add_tokens(["[BEGIN]"])
                tokenizer.eos_token = "[END]"
        else:
            raise

    return (added, tokenizer, info)


def get_model_path(model_id: str):
    # we should plausibly eventually remove current circ_models/ and replace with circ_models2/
    from rust_circuit.rrfs import RRFS_DIR

    return f"{RRFS_DIR}/circ_models2/{model_id}.circ"


def load_model_id(model_id: str, parser: Optional[Parser] = None) -> tuple[dict[str, Circuit], Any, TransformerInfo]:
    """Returns (circs, tokenizer, info)"""
    with open(get_model_path(model_id)) as f:
        return load_transformer_model_string(f.read(), parser=parser)


token_upcast_module = make_spec(
    P(
        """
'upcast_toks.sub' Add
  'upcast_toks' GeneralFunction cast_from_{device:None,dtype:None}_to_{device:None,dtype:int64}
    'upcast_toks.int16_toks' [] Symbol
  'upcast_toks.signed_int16_min' [] Scalar 32768 # int16 min, torch doesn't support uint16 :(
"""
    ),
    [P("'upcast_toks.int16_toks' [] Symbol")],
)



# imports {{{

# %%

from __future__ import annotations

import pytest

from rust_circuit import Index, Matcher, Parser, PrintOptions, default_index_traversal, push_down_index, restrict
from rust_circuit.py_utils import I

# %%

# }}}
# setup {{{

# %% [markdown]

# ## Setup

# %%

circuit = Parser(tensors_as_random=True, on_repeat_check_info_same=False, allow_hash_with_random=True)(
    """
0 'log_probs' [57, 50259] GeneralFunction log_softmax
  1 'logits' [57, 50259] Einsum ab,cb->ac
    2 'final.n' [57, 256] Add
      3 'final.n.w.bias' [256] Array 151e0290b45764bf65fe8e61
      4 'final.n.y_out' [57, 256] Einsum ab,a->ab
        5 'final.n.y_scale' [57, 256] Einsum ab,b->ab
          6 'final.n.z_mean' [57, 256] Einsum ab,bc->ac
            7 'final.inp' [57, 256] Add
              8 'tok_embeds' [57, 256] Array f738ed95c0f778275c1ecc91
              9 'a0' [57, 256] Einsum abc,bcd->ad
                10 'a0.comb_v' [57, 8, 32] Einsum abc,acd->bad
                  11 'a0.probs' [8, 57, 57] GeneralFunction softmax
                    12 'a0.scores' [8, 57, 57] Add
                      13 'a0.scores_mul_mask' [8, 57, 57] Einsum abc,bc->abc
                        14 'a0.scores_not_masked' [8, 57, 57] Einsum abc,adc,->abd
                          15 'a0.f_q' [8, 57, 32] Add
                            16 'a0.q' [8, 57, 32] Einsum abc,dc->adb
                              17 'a0.w.q' [8, 32, 256] Array ef8f32f62ccaa15af90ed096
                              18 'a0.n' [57, 256] Add
                                19 'a0.n.w.bias' [256] Array 925f1bb4ca52e7573018ba34
                                20 'a0.n.y_out' [57, 256] Einsum ab,a->ab
                                  21 'a0.n.y_scale' [57, 256] Einsum ab,b->ab
                                    22 'a0.n.z_mean' [57, 256] Einsum ab,bc->ac
                                      8 tok_embeds
                                      23 'a0.n.c.sub_mean' [256, 256] Array 4d11427c8d9c09763ca3c7de
                                    24 'a0.n.w.scale' [256] Array 03798fe8f36bb9e6271f8bcd
                                  25 'a0.n.full_mul' [57] GeneralFunction rsqrt
                                    26 'a0.n.c.var_p_eps' [57] Add
                                      27 'a0.n.var' [57] Einsum ab,ab,->a
                                        22 a0.n.z_mean
                                        22 a0.n.z_mean
                                        28 'a0.n.c.recip_h_size' [] Scalar 0.00390625
                                      29 'a0.n.eps' [] Scalar 0.00001
                            30 'a0.w.pos_q' [8, 57, 32] Einsum abc,dc->adb
                              17 a0.w.q
                              31 'w.pos_embeds' [57, 256] Array 47eda059261d25d4d7283df3
                          32 'a0.f_k' [8, 57, 32] Add
                            33 'a0.k' [8, 57, 32] Einsum abc,dc->adb
                              34 'a0.w.k' [8, 32, 256] Array 3eae23cb6134604792918d08
                              18 a0.n
                            35 'a0.w.pos_k' [8, 57, 32] Einsum abc,dc->adb
                              34 a0.w.k
                              31 w.pos_embeds
                          36 'a0.c.div_head_size' [] Scalar 0.17677669529663687
                        37 'a0.c.score_mask' [57, 57] Array 4d251ae1987ead35d704cd1b
                      38 'a0.c.score_neg_inf_bias' [57, 57] Array 3346b3e97cda641e9cbec006
                  39 'a0.v' [8, 57, 32] Einsum abc,dc->adb
                    40 'a0.w.v' [8, 32, 256] Array 9c803a30d26e5d50b28376dd
                    18 a0.n
                41 'a0.w.out' [8, 32, 256] Array a737a4b485f2e1cc3db22333
              42 'm0' [57, 256] Einsum ab,cb->ac
                43 'm0.act' [57, 1024] GeneralFunction gelu
                  44 'm0.add0' [57, 1024] Add
                    45 'm0.before_product0' [57, 1024] Einsum ab,cb->ac
                      46 'm0.n' [57, 256] Add
                        47 'm0.n.w.bias' [256] Array 9ae61c8299e708d1b3c12c75
                        48 'm0.n.y_out' [57, 256] Einsum ab,a->ab
                          49 'm0.n.y_scale' [57, 256] Einsum ab,b->ab
                            50 'm0.n.z_mean' [57, 256] Einsum ab,bc->ac
                              51 'm0.inp' [57, 256] Add
                                8 tok_embeds
                                9 a0
                              52 'm0.n.c.sub_mean' [256, 256] Array 4d11427c8d9c09763ca3c7de
                            53 'm0.n.w.scale' [256] Array c0b8ebacfbd81ed1c905e5d1
                          54 'm0.n.full_mul' [57] GeneralFunction rsqrt
                            55 'm0.n.c.var_p_eps' [57] Add
                              56 'm0.n.var' [57] Einsum ab,ab,->a
                                50 m0.n.z_mean
                                50 m0.n.z_mean
                                57 'm0.n.c.recip_h_size' [] Scalar 0.00390625
                              58 'm0.n.eps' [] Scalar 0.00001
                      59 'm0.w.w0' [1024, 256] Array dc59d55cbe737267d1acbbf9
                    60 'm0.w.b0' [1024] Array 7dbf23f05f2ff50ab6ae3d30
                61 'm0.w.w1' [256, 1024] Array 5002910549bc78a47941545c
              62 'm0.w.b1' [256] Array 67e1c57f81b33d2e6f6cfbe3
              63 'a1' [57, 256] Einsum abc,bcd->ad
                64 'a1.comb_v' [57, 8, 32] Einsum abc,acd->bad
                  65 'a1.probs' [8, 57, 57] GeneralFunction softmax
                    66 'a1.scores' [8, 57, 57] Add
                      67 'a1.scores_mul_mask' [8, 57, 57] Einsum abc,bc->abc
                        68 'a1.scores_not_masked' [8, 57, 57] Einsum abc,adc,->abd
                          69 'a1.f_q' [8, 57, 32] Add
                            70 'a1.q' [8, 57, 32] Einsum abc,dc->adb
                              71 'a1.w.q' [8, 32, 256] Array 91124557e0f6f11619a770d7
                              72 'a1.n' [57, 256] Add
                                73 'a1.n.w.bias' [256] Array 9c4b4d2ec8664cb710486d99
                                74 'a1.n.y_out' [57, 256] Einsum ab,a->ab
                                  75 'a1.n.y_scale' [57, 256] Einsum ab,b->ab
                                    76 'a1.n.z_mean' [57, 256] Einsum ab,bc->ac
                                      77 'a1.inp' [57, 256] Add
                                        8 tok_embeds
                                        9 a0
                                        42 m0
                                        62 m0.w.b1
                                      78 'a1.n.c.sub_mean' [256, 256] Array 4d11427c8d9c09763ca3c7de
                                    79 'a1.n.w.scale' [256] Array 0fcb1f000bb89dda2e86c04f
                                  80 'a1.n.full_mul' [57] GeneralFunction rsqrt
                                    81 'a1.n.c.var_p_eps' [57] Add
                                      82 'a1.n.var' [57] Einsum ab,ab,->a
                                        76 a1.n.z_mean
                                        76 a1.n.z_mean
                                        83 'a1.n.c.recip_h_size' [] Scalar 0.00390625
                                      84 'a1.n.eps' [] Scalar 0.00001
                            85 'a1.w.pos_q' [8, 57, 32] Einsum abc,dc->adb
                              71 a1.w.q
                              31 w.pos_embeds
                          86 'a1.f_k' [8, 57, 32] Add
                            87 'a1.k' [8, 57, 32] Einsum abc,dc->adb
                              88 'a1.w.k' [8, 32, 256] Array 15f4e9cd023913a9a854a514
                              72 a1.n
                            89 'a1.w.pos_k' [8, 57, 32] Einsum abc,dc->adb
                              88 a1.w.k
                              31 w.pos_embeds
                          90 'a1.c.div_head_size' [] Scalar 0.17677669529663687
                        91 'a1.c.score_mask' [57, 57] Array 4d251ae1987ead35d704cd1b
                      92 'a1.c.score_neg_inf_bias' [57, 57] Array 3346b3e97cda641e9cbec006
                  93 'a1.v' [8, 57, 32] Einsum abc,dc->adb
                    94 'a1.w.v' [8, 32, 256] Array 5aaef038d2194d529aea8758
                    72 a1.n
                95 'a1.w.out' [8, 32, 256] Array 4bfcebee9a186581d8de84cc
              96 'm1' [57, 256] Einsum ab,cb->ac
                97 'm1.act' [57, 1024] GeneralFunction gelu
                  98 'm1.add0' [57, 1024] Add
                    99 'm1.before_product0' [57, 1024] Einsum ab,cb->ac
                      100 'm1.n' [57, 256] Add
                        101 'm1.n.w.bias' [256] Array c8ee8c692edc6d92e7c2c219
                        102 'm1.n.y_out' [57, 256] Einsum ab,a->ab
                          103 'm1.n.y_scale' [57, 256] Einsum ab,b->ab
                            104 'm1.n.z_mean' [57, 256] Einsum ab,bc->ac
                              105 'm1.inp' [57, 256] Add
                                8 tok_embeds
                                9 a0
                                42 m0
                                62 m0.w.b1
                                63 a1
                              106 'm1.n.c.sub_mean' [256, 256] Array 4d11427c8d9c09763ca3c7de
                            107 'm1.n.w.scale' [256] Array 617a80d2cb0cb45318ef9885
                          108 'm1.n.full_mul' [57] GeneralFunction rsqrt
                            109 'm1.n.c.var_p_eps' [57] Add
                              110 'm1.n.var' [57] Einsum ab,ab,->a
                                104 m1.n.z_mean
                                104 m1.n.z_mean
                                111 'm1.n.c.recip_h_size' [] Scalar 0.00390625
                              112 'm1.n.eps' [] Scalar 0.00001
                      113 'm1.w.w0' [1024, 256] Array c46ef0b6b06fa0a25a9e8f74
                    114 'm1.w.b0' [1024] Array 1eeee3393a8ef176b116cec0
                115 'm1.w.w1' [256, 1024] Array 040df31a221c659c1da228b7
              116 'm1.w.b1' [256] Array 7010bd45175fc579786a025a
            117 'final.n.c.sub_mean' [256, 256] Array 4d11427c8d9c09763ca3c7de
          118 'final.n.w.scale' [256] Array 3900ff7658d3f069ae29926e
        119 'final.n.full_mul' [57] GeneralFunction rsqrt
          120 'final.n.c.var_p_eps' [57] Add
            121 'final.n.var' [57] Einsum ab,ab,->a
              6 final.n.z_mean
              6 final.n.z_mean
              122 'final.n.c.recip_h_size' [] Scalar 0.00390625
            123 'final.n.eps' [] Scalar 0.00001
    124 'w.unembed' [50259, 256] Array 460a894d3c021fb94cabdfb1
"""
)

# %%
# }}}
# %%

pushed = push_down_index(
    Index(circuit, [17]), restrict(default_index_traversal(), term_early_at="final.inp"), suffix="_idx"
)
pushed.print(PrintOptions(shape_only_when_necessary=False))

# %%

pushed_many = push_down_index(
    Index(circuit, I[2:7]), restrict(default_index_traversal(), term_early_at="final.inp"), suffix="_idx"
)
pushed_many.print(PrintOptions(shape_only_when_necessary=False))

# %%

pushed_sub = push_down_index(Index(pushed_many, I[1:3]), default_index_traversal(), suffix="_IDX")
pushed_sub.print(PrintOptions(shape_only_when_necessary=False))

# %%

pushed_far = push_down_index(Index(circuit, I[2:7]), suffix="_idx")  # push down all the way!
pushed_far.print(PrintOptions(shape_only_when_necessary=False))

_ = Matcher("m1_idx").get_unique(pushed_far)
with pytest.raises(RuntimeError):
    Matcher("m1").get_unique(pushed_far)  # this doesn't exist
_ = Matcher("a1_idx").get_unique(pushed_far)
with pytest.raises(RuntimeError):
    Matcher("a1").get_unique(pushed_far)  # this doesn't exist
# but both of these will exist due to sequence position translation from a1
_ = Matcher("m0_idx").get_unique(pushed_far)
_ = Matcher("m0").get_unique(pushed_far)
_ = Matcher("a0_idx").get_unique(pushed_far)
_ = Matcher("a0").get_unique(pushed_far)



from typing import cast

import hypothesis
import torch
from torch.testing import assert_close

import rust_circuit as rc
from interp.circuit.testing.utils import mark_not_interesting_if
from rust_circuit import *

from .test_rust_rewrite import get_c_st

po = rc.PrintOptions()  # (tensor_index_literal=True)#, bijection=False)


@hypothesis.settings(deadline=None)
@hypothesis.given(get_c_st(max_growth_steps=20))
@mark_not_interesting_if(SchedulingOOMError)
def test_schedule(circ):
    raw_test_schedule(circ)


def raw_test_schedule(circ):
    circ = rc.cast_circuit(circ, TorchDeviceDtypeOp(device="cpu", dtype="float64"))
    # circ = rc.substitute_all_modules(circ)
    schedule = optimize_to_schedule(circ)
    # hypothesis.note("circ: " + circ.repr(po))
    # hypothesis.note("optimized: " + rc.optimize_circuit(circ).repr(po))

    hypothesis.assume(
        all(not ix.cast_index().has_tensor_axes() for ix in circ.get(rc.Index))
    )  # might go away due to Index.canonicalize, so can't replace them

    circ_out = circ.evaluate()
    assert_close(circ_out, schedule.evaluate().type(circ_out.dtype))

    circ_replaced = circ.update(rc.Array, lambda x: Array(cast(Array, x).value + 9.1))
    # hypothesis.note("replaced: " + circ_replaced.repr(po))
    # hypothesis.note("replaced optimized: " + rc.optimize_circuit(circ_replaced).repr(po))
    # hypothesis.note(repr({x.hash.hex(): x for x in rc.all_children(circ) if x.is_array()}))
    # hypothesis.note(repr({x.hash.hex(): x for x in schedule.constants.values()}))
    circ_replaced_in_rust = schedule.replace_tensors(
        {x.hash: x.cast_array().value + 9.1 for x in rc.all_children(circ) if x.is_array()},
        allow_missing=True,  # sometimes Arrays go away, e.g. due to 0*x = 0
    )
    # hypothesis.note(repr(schedule))
    # hypothesis.note(repr([(x, y, y.hash.hex()) for x, y in schedule.constants.items()]))
    x = circ_replaced.evaluate()
    y = circ_replaced_in_rust.evaluate().type(x.dtype)
    # hypothesis.note(repr(x))
    # hypothesis.note(repr(y))
    assert_close(x, y)



import math
import urllib.parse
import webbrowser
from json import dumps
from typing import Any, Callable, Dict, Optional

import rust_circuit as rc
from rust_circuit import Circuit, PrintOptions

ui_default_hidden_matcher = rc.Matcher(rc.Scalar, rc.Tag)
ui_no_ln_matcher = rc.Matcher("ln", rc.Matcher.regex(r"\.*ln\.*"))


def circuit_to_json(circ: Circuit) -> Any:
    graph = {}
    printer = PrintOptions()

    def visitor(circ: Circuit):
        childIndices: Dict[str, Any] = {}
        extra_info = printer.repr_extra_info(circ)
        for i, child in enumerate(circ.children):
            to_append = {"i": i, "n": ""}
            if circ.is_einsum():
                to_append["n"] = extra_info.split("->")[0].split(",")[i]
            if circ.is_module():
                circ_mod = circ.cast_module()
                if i == 0:
                    to_append["n"] = "spec"
                else:
                    to_append["n"] = circ_mod.spec.arg_specs[i - 1].symbol.name

            if child.hash_base16 in childIndices:
                childIndices[child.hash_base16].append(to_append)
            else:
                childIndices[child.hash_base16] = [to_append]
        graph[circ.hash_base16] = {
            "hash": circ.hash_base16,
            "childIndices": childIndices,
            "meta": {
                "name": circ.op_name,
                "shape": [None if x == rc.Size.NONE else x for x in circ.shape],
                "numel": rc.oom_fmt(circ.numel),
                "kind": circ.__class__.__name__,
                "extra": extra_info,
            },
        }

    circ.visit(visitor)
    return graph


def eval_code_on_circuit(circ_string: str, code: str) -> Any:
    """Big security vulnerability if called remotely"""
    circ = rc.Parser(
        tensors_as_random=True, tensors_as_random_device_dtype=rc.TorchDeviceDtypeOp("cpu", "float32")
    ).parse_circuit(circ_string)
    return eval(code, {"circ": circ, "rc": rc}, {})


def circuit_graph_ui(
    circ: Circuit,
    *,
    base_url: str = "http://interp-tools.redwoodresearch.org",
    annotators: dict[str, Callable[[Circuit], str]] = {},
    enabled_annotators: list[str] = [],
    default_hidden: Optional[set[Circuit]] = None,
    default_shown: Optional[set[Circuit]] = None,
):
    if default_hidden is not None and default_shown is not None:
        raise Exception("only one of default_hidden and default_shown can be set")
    if default_hidden is None:
        default_hidden = set()
    if default_shown is not None:

        def fny(c: Circuit):
            assert isinstance(default_shown, set)  # for linter
            assert isinstance(default_hidden, set)  # for linter
            if c not in default_shown:
                default_hidden.add(c)

        circ.visit(fny)
    default_hidden.discard(circ)  # root can't be hidden
    graph = circuit_to_json(circ)
    data = {
        "graph": graph,
        "rootChildHashes": [circ.hash_base16],
        "annotators": list(annotators.keys()),
        "enabledAnnotators": enabled_annotators,
        "default_hidden": [x.hash_base16 for x in default_hidden],
    }
    url = base_url + "/#/layout?data=" + urllib.parse.quote(dumps(data))
    print(url)
    webbrowser.open(url)



from __future__ import annotations

import functools
import uuid
from collections import Counter
from typing import Callable, Dict, Hashable, List, Optional, Protocol, Set, Tuple, Union
from warnings import warn

import attrs
import torch

from rust_circuit import (
    Circuit,
    GeneralFunction,
    GeneralFunctionShapeInfo,
    GeneralFunctionSpecBase,
    IterativeMatcher,
    IterativeMatcherIn,
    MatcherIn,
    PrintHtmlOptions,
    PrintOptions,
    hash_tensor,
    new_traversal,
    restrict,
)

from .dataset import Dataset, color_dataset


@attrs.define(eq=True, hash=True, init=False)
class PoolAnnotation:
    """
    Pool annotation for a cond sampler. See pool.py demo for more info.

    size: size of the pool, i.e. how many different datasets will be sampled. A pool size of 1 has semantic meaning
        (you want to enforce things being sampled together). A pool of size, say, 4-10 is useful for performance.
        Intermediate sizes seem confusing, it's not clear to me why you would want them.
    id: to define which annotations, and therefore which cond samplers, are sharing a pool.
    """

    size: int
    id: uuid.UUID

    def __init__(self, size: int = 4, id: Optional[uuid.UUID] = None):
        if size <= 0:
            raise ValueError(size)
        self.size = size
        self.id = uuid.uuid4() if id is None else id


@attrs.define(hash=True, eq=True)
class PoolNotPresent:
    ...


PNP = PoolNotPresent()
MaybePoolAnnotation = Union[PoolAnnotation, PoolNotPresent]

DsEqClassT = Hashable
Pool = List[Optional[Dataset]]
PoolPerDsEqClass = dict[DsEqClassT, Pool]
SharedPoolKey = Tuple["CondSampler", Dataset]
PoolsPerSampler = dict[SharedPoolKey, PoolPerDsEqClass]


@attrs.define(hash=True, eq=True)
class CondSampler(Protocol):
    """
    Responsible for sampling a new dataset given a source dataset to draw from, and a reference dataset.

    Parameters:
    'pool_annot`: See pool.py demo for more info. If PoolNotPresent (the default), no sample reuse will
        happen. Using pools is recommended if your treeified model is too big (e.g for deep networks with
        non-trivial samplers).
    """

    pool_annot: MaybePoolAnnotation = PNP

    def __call__(self, ref: Dataset, ds: Dataset, rng=None) -> Dataset:
        ...

    def ds_eq_class(self, ds: Dataset) -> Hashable:
        """
        Defines what reference datasets share a pool. By default, no sharing is allowed.

        This should return some value which will be the same for datasets that "are the same, according to this sampler."

        Note: not guaranteed to actually agree with __call__, that's up to the author.
        """
        return ds

    def sample_and_update_pool(
        self, ref: Dataset, ds: Dataset, rng: torch.Generator, pools_per_sampler: PoolsPerSampler
    ) -> Dataset:
        """
        Reuses a dataset from the pool or samples a new one and updates the pool.
        """
        if isinstance(self.pool_annot, PoolNotPresent):
            return self(ref, ds, rng)

        # Get pools for this sampler. Based on hash, which depends on its PoolAnnotation as well as other attrs.
        if (self, ds) not in pools_per_sampler:
            pools_per_sampler[(self, ds)] = PoolPerDsEqClass()
        pools = pools_per_sampler[(self, ds)]

        # Get the pool for the ref ds. Depends on what ref ds's eq class is, according to this sampler.
        ref_eq_class = self.ds_eq_class(ref)
        if ref_eq_class not in pools:
            pools[ref_eq_class] = [None for _ in range(self.pool_annot.size)]
        pool = pools[ref_eq_class]

        # Pick a ds from the pool, lazily sampling a new one if needed.
        picked = torch.randint(low=0, high=len(pool), size=(), generator=rng)
        picked_ds = pool[picked]
        if picked_ds is None:
            picked_ds = self(ref, ds, rng)
            pool[picked] = picked_ds
        return picked_ds

    def pretty_str(self, data: Optional[Dataset], datum_idx=0) -> str:
        # Not __str__ because it takes a dataset
        if data is None:
            extra = ""
        else:
            datum = data[datum_idx]
            extra = f"d={datum}"
            cond_extra = self.str_cond(datum)
            if cond_extra:
                extra += ", " + cond_extra
            extra = f"({extra})"
        return f"{self.class_str()}{extra}"

    def class_str(self):
        return f"{self.__class__.__name__}" + (f"#{hash(self.pool_annot) & 1023}" if self.pool_annot != PNP else "")

    def str_cond(self, datum: Dataset) -> str:
        return ""


@attrs.define(hash=True, eq=True, init=False)
class FuncSampler(CondSampler):
    """
    Samples each datum uniformly from the subset of ds that agrees on the value of func. That is, for x in ref_ds we sample a corresponding x' uniformly from `[y for y in ds if func(y) == func(x)]`.

    This impliments the standard causal scrubbing sampling operation as described in the writeup, if func returns the feature(s) computed by the interp node.

    `func: Dataset -> Tensor` must return a 1 or 2 dimensional tensor, where the first dimesion is the same length as the dataset the function is called on.
    """

    func: Callable[[Dataset], torch.Tensor] = attrs.field(
        kw_only=True
    )  # irrelevant since we are writing our own init method, but attrs complains otherwise

    def __init__(
        self,
        func: Callable[[Dataset], torch.Tensor],
        pool_annot: MaybePoolAnnotation = PNP,
    ):
        self.func = func
        self.pool_annot = pool_annot

    @staticmethod
    @functools.lru_cache(maxsize=128)
    def get_matching_idxs(
        func: Callable[[Dataset], torch.Tensor], ref: Dataset, ds: Dataset
    ) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        r_vals = func(ref)
        d_vals = func(ds)
        r_vals = r_vals[:, None] if r_vals.ndim == 1 else r_vals
        d_vals = d_vals[:, None] if d_vals.ndim == 1 else d_vals
        result = []
        for elem in torch.unique(r_vals, dim=0):
            matching_idxs_r = torch.nonzero(torch.all(r_vals == elem, dim=1).ravel(), as_tuple=True)[0]
            matching_idxs_d = torch.nonzero(torch.all(d_vals == elem, dim=1).ravel(), as_tuple=True)[0]
            r_count = len(matching_idxs_r)
            d_count = len(matching_idxs_d)

            assert d_count > 0, f"no matching idxs found for value {elem} of condition {func}"
            if (
                d_count < r_count
            ):  # Unfortunately happens a lot when func is injective, because r may contain duplicates but d won't
                warn(f"not enough unique dataset samples for value {elem} of condition {func}, {d_count} < {r_count}")
            result.append((matching_idxs_r, matching_idxs_d))
        return result

    def __call__(self, ref: Dataset, ds: Dataset, rng=None) -> Dataset:
        r_d_pairs = FuncSampler.get_matching_idxs(self.func, ref, ds)
        idxs = torch.full((len(ref),), -1, dtype=torch.int64, device=r_d_pairs[0][0].device)
        for matching_idxs_r, matching_idxs_d in r_d_pairs:
            idxs[matching_idxs_r] = matching_idxs_d[
                torch.multinomial(
                    torch.ones(len(matching_idxs_d), device=matching_idxs_d.device, dtype=torch.float32),
                    num_samples=len(matching_idxs_r),
                    replacement=True,
                    generator=rng,
                )
            ]

        assert (idxs != -1).all(), "this should never happen!"
        return ds[idxs]

    def ds_eq_class(self, ds):
        return hash_tensor(self.func(ds))

    def str_cond(self, datum: Dataset) -> str:
        out = self.func(datum)[0]
        if isinstance(out, torch.Tensor) and out.numel() == 1:
            out = out.item()  # type: ignore
        return f"f(d)={str(out)}"


@attrs.define(hash=True, eq=True)
class UncondSampler(CondSampler):
    """
    Samples randomly without conditioning on the reference dataset.
    """

    def __call__(self, ref: Dataset, ds: Dataset, rng=None) -> Dataset:
        return ds.sample(len(ref), rng)

    def ds_eq_class(self, ds):
        return ()


@attrs.define(hash=True, eq=True)
class UncondTogetherSampler(UncondSampler):
    """
    Samples randomly without conditioning on the reference dataset,
    and has a pool of size 1 so a single sample will be reused.

    By default all instances of this class are equal, so all nodes with an
    UncondTogetherSampler will get the same ds, but you can pass a different
    uuid at init if you want to create groups of nodes.
    """

    def __init__(self, id=uuid.UUID("566ae005-4a34-4e37-9383-731e1a722ef2")):
        self.pool_annot = PoolAnnotation(1, id=id)


@attrs.define(hash=True, eq=True)
class ExactSampler(CondSampler):
    def __call__(self, ref: Dataset, ds: Dataset, rng=None) -> Dataset:
        return ref  # should maybe check this is valid draw from ds


@attrs.define(hash=True, eq=True)
class FixedOtherSampler(UncondSampler):
    other: Dataset = attrs.field(
        kw_only=True
    )  # irrelevant since we are writing our own init method, but attrs complains otherwise

    def __init__(self, other: Dataset):
        super().__init__()
        self.other = other

    def __call__(self, ref: Dataset, ds: Dataset, rng=None) -> Dataset:
        assert len(self.other) == len(ref), (len(self.other), len(ref))
        return self.other  # should maybe check this is a valid draw from ds


def chain_excluding(parent: IterativeMatcher, child: IterativeMatcherIn, term_early_at: MatcherIn = False):
    """Matches `child` from `parent`, excluding any intermediate nodes matched by `term_early_at`."""
    if term_early_at is False:
        return parent.chain(child)
    return restrict(parent.chain(IterativeMatcher(child)), term_early_at=term_early_at)


@attrs.define(frozen=True)
class SampledInputs:
    """Holds the sampled inputs at each node of an interpretation graph."""

    datasets: dict[InterpNode, Dataset] = attrs.field(factory=dict)
    other_inputs_datasets: dict[InterpNode, Dataset] = attrs.field(factory=dict)
    sampler_pools: PoolsPerSampler = attrs.field(factory=dict)

    def __getitem__(self, node: InterpNode) -> Dataset:
        if node.is_leaf():
            return self.datasets[node]
        else:
            return self.other_inputs_datasets[node]

    def get(self, node: InterpNode, default: Optional[Dataset] = None) -> Optional[Dataset]:
        try:
            return self[node]
        except KeyError:
            return default


class InterpNodeGeneralFunctionSpec(GeneralFunctionSpecBase):
    def __init__(self, function, name):
        self.function = function
        self.name = name

    def name(self):
        return self.name

    def function(self, *tensors):
        return self.function(*tensors)

    def get_shape_info(self, *shapes):
        # empty shape
        return GeneralFunctionShapeInfo((), 0, [False] * len(shapes))


class InterpNode:
    """
    Interpretation graph (tree!) node.

    It doesn't have to be a tree in general, but we are lazy and haven't written treeification code for it.

    cond_sampler: samples data that this node is "indifferent between", e.g.
        data agreeing on a feature with a reference datum if this node only cares about that feature
    name: should be unique within an interpretation. it's nice to be able to uniquely identify things
        by name rather than by more opaque id.
    chidren: interpretation graph nodes that are children of this one. this is used for recursive sampling
        when causal scrubbing, but not much else. it's also kind of nice to think of your interpretation
        as being a graph.
    other_inputs_sampler: formally, causal scrubbing hypotheses should have graph isomorphisms between
        the interpretation and the graph you are trying to explain. for convenience, you can instead specify
        how inputs into the image of this node should be sampled, if not otherwise stated. this is equivalent
        to adding children to this node to make it surjective, and using this sampler as the cond_sampler for
        all of those. not applicable for leaves. (for example, you could have your interpretation just consist
        of the circuit you think is important, and say that other inputs should be sampled randomly).
    """

    def __init__(self, cond_sampler: CondSampler, name: str, other_inputs_sampler: CondSampler = UncondSampler()):
        self._name = name
        self.cond_sampler = cond_sampler
        self.other_inputs_sampler = other_inputs_sampler
        self._children: Tuple["InterpNode", ...] = tuple()

    @property
    def name(self) -> str:
        return self._name

    @property
    def children(self) -> Tuple["InterpNode", ...]:
        return self._children

    def make_descendant(
        self,
        cond_sampler: CondSampler,
        name: str,
        other_inputs_sampler: CondSampler = UncondSampler(),
    ) -> "InterpNode":
        """Create a child node and add it to this node's children."""
        d = self.__class__(name=name, cond_sampler=cond_sampler, other_inputs_sampler=other_inputs_sampler)
        self._children = self.children + (d,)
        return d

    def is_leaf(self) -> bool:
        return len(self.children) == 0

    def get_tree_size(self) -> int:
        """Gets the number of nodes in the interp graph tree rooted at this node."""
        return 1 + sum([c.get_tree_size() for c in self.children])

    def get_nodes(self: "InterpNode") -> List["InterpNode"]:
        return [
            self,
        ] + [n for c in self.children for n in c.get_nodes()]

    def get_descendants_with_parents(self: "InterpNode") -> List[Tuple["InterpNode", "InterpNode"]]:
        return [(self, c) for c in self.children] + [
            pair for c in self.children for pair in c.get_descendants_with_parents()
        ]

    def print(
        self,
        options: Optional[Union[PrintOptions, PrintHtmlOptions]] = None,
        sampled_inputs: SampledInputs = SampledInputs(),
        print_data=True,
        color_by_data=True,
        repr=False,
        datum_idx=0,
    ):
        """
        Parameters:
            print_data: if True, prints a sample datum from cond_sampler as a comment
            color_by_data: if True, colors nodes by ds
            repr: if True, returns string instead of printing it (used for testing)
        """

        def to_circuit(i: InterpNode) -> Tuple[Circuit, Dict[Circuit, "InterpNode"]]:
            node_circuit_map = {}
            child_circuits = []
            for child in i.children:
                child_circuit, child_map = to_circuit(child)
                for k, v in child_map.items():
                    node_circuit_map[k] = v
                child_circuits.append(child_circuit)

            spec = InterpNodeGeneralFunctionSpec(lambda x: x, "")
            this_circuit = GeneralFunction(*child_circuits, spec=spec, name=i.name)
            node_circuit_map[this_circuit] = i
            return this_circuit, node_circuit_map

        c, map = to_circuit(self)

        if options is None:
            new_options: Union[PrintOptions, PrintHtmlOptions] = PrintHtmlOptions(traversal=new_traversal())
        else:
            new_options = options.evolve()
        if print_data:
            new_options.commenters += [
                lambda c: map[c].str_samplers(sampled_inputs=sampled_inputs, datum_idx=datum_idx)
            ]
        if isinstance(new_options, PrintOptions):
            new_options.bijection = False
        if color_by_data:
            new_options.colorer = lambda c: color_dataset(  # type: ignore
                sampled_inputs.datasets.get(map[c]),
                html=isinstance(new_options, PrintHtmlOptions),
            )

        if repr:
            assert isinstance(new_options, PrintOptions)
            return c.repr(new_options)
        else:
            c.print(new_options)

    def _sample_into(
        self,
        rng,
        source_ds,
        parent_ds,
        into: SampledInputs,
        recursive: bool = True,
    ):
        # The one place that mutating SampledInputs is allowed
        ds = self.cond_sampler.sample_and_update_pool(parent_ds, source_ds, rng, into.sampler_pools)
        if not self.is_leaf():
            into.other_inputs_datasets[self] = self.other_inputs_sampler.sample_and_update_pool(
                ds, source_ds, rng, into.sampler_pools
            )
        into.datasets[self] = ds
        if recursive:
            for child in self.children:
                child._sample_into(rng, source_ds, ds, into)

    def sample(
        self,
        rng,
        source_ds,
        parent_ds,
        recursive=True,
    ) -> SampledInputs:
        into = SampledInputs()
        self._sample_into(rng, source_ds, parent_ds, into, recursive=recursive)
        return into

    def str_samplers(self, sampled_inputs: SampledInputs, datum_idx=0) -> str:
        cond_sampler_str = self.cond_sampler.pretty_str(sampled_inputs.datasets.get(self), datum_idx=datum_idx)
        samplers_str = f"cond_sampler={cond_sampler_str}"
        if not self.is_leaf():
            other_sampler_str = self.other_inputs_sampler.pretty_str(
                sampled_inputs.other_inputs_datasets.get(self), datum_idx=datum_idx
            )
            samplers_str += f", other_inputs_sampler={other_sampler_str}"
        return samplers_str

    def pretty_str(self, sampled_inputs: SampledInputs) -> str:
        return f"InterpNode(name={self.name}, {self.str_samplers(sampled_inputs)}, children={[c.name for c in self.children]})"

    def __str__(self):
        return self.pretty_str(SampledInputs())

    def __repr__(self):
        # Not the best repr; but this makes errors nicer for pyo stuff (which will repr rather than str)
        return str(self)


corr_root_matcher = restrict(new_traversal(), term_if_matches=True)


class IllegalCorrespondenceError(Exception):
    ...


DatasetOrInputNames = Union[Dataset, Set[str]]


def get_input_names(ds: DatasetOrInputNames) -> Set[str]:
    return ds.input_names if isinstance(ds, Dataset) else ds


def to_inputs(matcher: IterativeMatcher, ds: DatasetOrInputNames) -> IterativeMatcher:
    return matcher.chain(get_input_names(ds))


class Correspondence:
    """
    Holds correspondences between interp graph and model.

    The parts of the model are pointed at by IterativeMatchers picking out subgraphs of the model (with a single source and sink).
    In theory land we imagine that our correspondences are surjective, but in practice this is painful to implement: you either
    need drastic circuit rewrites, or very many InterpNodes and entries in your correspondence. To make life easier, we allow
    correspondences to model "branches" picked out by matchers.

    corr: mapping from interpretation to branches in the model
    """

    corr: Dict[InterpNode, IterativeMatcher]
    i_names: Dict[str, InterpNode]

    def __init__(self):
        self.corr = {}
        self.i_names = {}

    def copy(self):
        """Shallow copy of a Correspondence. The InterpNodes and matchers are attached to a new Correspondence, but
        not themselves copied."""
        new = Correspondence()
        new.corr = self.corr.copy()  # type: ignore
        new.i_names = self.i_names.copy()
        return new

    def __len__(self):
        return len(self.corr)

    def __getitem__(self, i: InterpNode) -> IterativeMatcher:
        return self.corr[i]

    def get_by_name(self, s: str) -> InterpNode:
        return self.i_names[s]

    def add(self, i_node: InterpNode, m_node: IterativeMatcher):
        if i_node in self.corr:
            raise ValueError("Node already in correspondence!", i_node, m_node)
        if i_node.name in self.i_names:
            raise ValueError(
                "Different node with same name already in correspondence!", self.i_names[i_node.name], i_node, m_node
            )
        self.corr[i_node] = m_node
        self.i_names[i_node.name] = i_node

    def replace(self, i_node: InterpNode, m_node: IterativeMatcher):
        if i_node not in self.corr:
            raise ValueError("Node not found in correspondence!", i_node, m_node)
        assert i_node.name in self.i_names, (
            "Node in corr but its name is not in names map, should never happen!",
            i_node,
            m_node,
        )
        self.corr[i_node] = m_node
        self.i_names[i_node.name] = i_node

    def get_root(self) -> InterpNode:
        if len(self) == 0:
            raise IllegalCorrespondenceError("Empty correspondence has no root!")

        i_nodes = list(self.corr.keys())

        # store parent counts and check that exactly one is 0
        parent_counts = {i: 0 for i in i_nodes}  # this doesn't implement hash or eq but we can just id it
        for i_node in i_nodes:
            for child in i_node.children:
                parent_counts[child] += 1
        roots = [i for i, count in parent_counts.items() if count == 0]
        if len(roots) > 1:
            raise IllegalCorrespondenceError("Found multiple roots!", roots)
        if len(roots) == 0:
            raise IllegalCorrespondenceError("Found no root, not a tree! (No node is an ancestor of all other nodes)")

        return roots[0]

    def in_dfs_order(self) -> List[Tuple[InterpNode, IterativeMatcher]]:
        if len(self) == 0:
            return []
        return [(i_node, self[i_node]) for i_node in self.get_root().get_nodes()]

    def in_dfs_order_with_parents(self) -> List[Tuple[Optional[InterpNode], InterpNode, IterativeMatcher]]:
        if len(self) == 0:
            return []
        root = self.get_root()
        root_w_parent: List[Tuple[Optional[InterpNode], InterpNode, IterativeMatcher]] = [(None, root, self[root])]
        return root_w_parent + [
            (i_node, i_child, self[i_child]) for (i_node, i_child) in root.get_descendants_with_parents()
        ]

    def check_complete(self):
        for i_node in self.corr.keys():
            children = (
                i_node.children
            )  # this used to be i_node.get_nodes (all descendents) but just .children is enough
            for d in children:
                if d not in self.corr:
                    raise IllegalCorrespondenceError(
                        "Node is a descendent of a node in the corr but is not in the corr itself",
                        d,
                        i_node,
                    )

    def check_well_defined(self, circuit: Circuit, ds: Optional[DatasetOrInputNames] = None):
        """
        Checks that each interp graph node matches exactly one circuit node.
        This may fail if e.g. your circuit has some "basically the same" nodes, like one with an extra 1 dim, but generally should pass.
        Includes checking the implicit input nodes exist (by extending matchers to the inputs).
        """
        for m_node in self.corr.values():
            try:
                m_node.get_unique(circuit)
            except RuntimeError:
                raise IllegalCorrespondenceError(
                    f"matcher {m_node} failed well defined check. This check is done in dfs order so all parents must have passed."
                )
            if ds is not None:
                m_to_inputs_matcher = to_inputs(m_node, ds)
                if not m_to_inputs_matcher.are_any_found(circuit):
                    raise IllegalCorrespondenceError(
                        f"matcher extended to inputs {m_to_inputs_matcher} failed existence check. This check is done in dfs order so all parents must have passed."
                    )

    def check_injective_on_treeified(self, circuit: Circuit):
        """
        Checks that every circuit node is matched by at most one interp graph node.
        We expect this to be true after the circuit is tree-ified but not before
        (and maybe not after we replace the treeified inputs, as some of them may
        turn out to be identical).
        """
        if len(self) == 0:
            return
        all_matched = self.get_all_matched(circuit)
        multiple_matches = [(c, i) for (c, i) in Counter(all_matched).items() if i > 1]
        if len(multiple_matches) != 0:
            raise IllegalCorrespondenceError(multiple_matches)

    def check_root_is_model(self):
        root_branch = self.corr[self.get_root()]
        if root_branch != corr_root_matcher:
            raise IllegalCorrespondenceError("the root of the model branch should be empty")

    def get_all_matched(self, circuit: Circuit) -> List[Circuit]:
        """Gets all circuit nodes matched by interp graph nodes."""
        return [c for m_node in self.corr.values() for c in m_node.get(circuit)]

    def check(self, circuit: Circuit, circuit_treeified=True):
        self.check_complete()
        self.check_root_is_model()
        if circuit_treeified:
            self.check_injective_on_treeified(circuit)

    def sample(self, source_ds, rng, ref_ds) -> SampledInputs:
        return self.get_root().sample(source_ds=source_ds, parent_ds=ref_ds, rng=rng)

    def __repr__(self):
        return f"{self.__class__.__name__}(corr={repr(self.corr)})"



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Union, TypeVar, Sequence, Literal

from dataclasses import dataclass

import rust_circuit.index_util.internal.parse as parse


@dataclass(frozen=True, eq=True)
class OutAxisIdx:
    axis: int


@dataclass(frozen=True, eq=True)
class IntLitIdx:
    value: int


@dataclass(frozen=True, eq=True)
class PosDataIdx:
    pass


if TYPE_CHECKING:
    T = TypeVar("T")


def is_unique(lst: list[T]) -> bool:
    return len(set(lst)) == len(lst)


OUT_NAME: str = "out"

if TYPE_CHECKING:
    IdxExpr = Union[OutAxisIdx, IntLitIdx, PosDataIdx]


@dataclass
class IdxRenderer:
    axis_names: list[str]
    pos_name: str
    pos_indices: list[IdxExpr]

    def render_indices(self, indices: list[IdxExpr], out: list[str]) -> None:
        if len(indices) > 0:
            out.append("[")
            for i, idx in enumerate(indices):
                if i > 0:
                    out.append(", ")
                self.render_idx(idx, out)
            out.append("]")

    def render_idx(self, idx: IdxExpr, out: list[str]) -> None:
        if isinstance(idx, OutAxisIdx):
            out.append(self.axis_names[idx.axis])
        elif isinstance(idx, IntLitIdx):
            out.append(str(idx.value))
        elif isinstance(idx, PosDataIdx):
            out.append(self.pos_name)
            self.render_indices(self.pos_indices, out)
        else:
            assert False


@dataclass(repr=False)
class GatherSpec:
    out_rank: int
    src_indices: list[IdxExpr]
    pos_indices: list[IdxExpr]

    axis_names: list[str]
    src_name: str
    pos_name: str

    def __post_init__(self) -> None:
        assert sum(1 for i in self.src_indices if isinstance(i, PosDataIdx)) == 1
        assert all(isinstance(i, OutAxisIdx) or isinstance(i, IntLitIdx) for i in self.pos_indices)
        assert all(i.axis < self.out_rank for i in self.src_indices + self.pos_indices if isinstance(i, OutAxisIdx))
        assert all(parse.is_ident(n) for n in self.axis_names)
        assert len(self.axis_names) == self.out_rank
        assert parse.is_ident(self.src_name)
        assert parse.is_ident(self.pos_name)
        assert is_unique(self.axis_names + [self.src_name, self.pos_name, OUT_NAME])

    def render(self, out: list[str]) -> None:
        renderer = IdxRenderer(axis_names=self.axis_names, pos_name=self.pos_name, pos_indices=self.pos_indices)
        out.append(f"{OUT_NAME}")
        renderer.render_indices([OutAxisIdx(i) for i in range(self.out_rank)], out)
        out.append(" = ")
        out.append(self.src_name)
        renderer.render_indices(self.src_indices, out)

    def __str__(self) -> str:
        out: list[str] = []
        self.render(out)
        return "".join(out)

    def __repr__(self) -> str:
        return f"GatherSpec.parse({str(self)!r}, {[self.src_name, self.pos_name]!r})"

    @staticmethod
    def resolve(equation: parse.Equation, arr_names: list[str]) -> "GatherSpec":
        if equation.kind != "=":
            raise ValueError(f"Gather formula must be written with '=' operator; got: {equation.kind!r}")

        if len(arr_names) != 2:
            raise ValueError(f"Expected exactly two input array names for gather; got: {arr_names!r}")

        lhs = equation.lhs
        if not (isinstance(lhs, parse.IdentExpr) and lhs.name == OUT_NAME):
            raise ValueError(f"Invalid left hand side; expected {OUT_NAME!r} array, got: {lhs}")

        if not is_unique([OUT_NAME] + arr_names):
            raise ValueError("Array names must be unique")

        forbidden_axis_names = set([OUT_NAME] + arr_names)
        axis_names = []
        for i, e in enumerate(lhs.idx):
            if not (isinstance(e, parse.IdentExpr) and e.idx == []):
                raise ValueError(f"Invalid index name: {e}")
            if e.name in forbidden_axis_names:
                raise ValueError(f"Index name conflicts with a name already in scope: {e.name!r}")
            forbidden_axis_names.add(e.name)
            axis_names.append(e.name)
        axis_name_indices = {n: i for i, n in enumerate(axis_names)}

        out_rank = len(axis_names)

        rhs = equation.rhs
        if not isinstance(rhs, parse.IdentExpr):
            raise ValueError(f"Invalid right hand side; expected array access expression, got: {rhs}")
        src_name = rhs.name
        if src_name not in arr_names:
            raise ValueError(f"Invalid 'src' array name; expected one of {arr_names!r}, got: {src_name!r}")

        pos_name = [n for n in arr_names if n != src_name][0]

        def resolve_direct_index(e: parse.Expr) -> Union[OutAxisIdx, IntLitIdx]:
            if isinstance(e, parse.IdentExpr) and e.idx == []:
                axis_idx = axis_name_indices.get(e.name)
                if axis_idx is None:
                    raise ValueError(f"Index name not in scope; expected one of {axis_names!r}, got: {e.name!r}")
                return OutAxisIdx(axis_idx)
            elif isinstance(e, parse.IntExpr):
                return IntLitIdx(e.value)
            else:
                raise ValueError(f"Invalid index expression: {e}")

        src_indices: list[IdxExpr] = []
        pos_indices: Optional[list[IdxExpr]] = None
        for i, e in enumerate(rhs.idx):
            if isinstance(e, parse.IdentExpr) and e.name == pos_name:
                src_indices.append(PosDataIdx())
                if pos_indices is not None:
                    raise ValueError(f"Index array {pos_name!r} appears more than once")
                pos_indices = [resolve_direct_index(e2) for e2 in e.idx]
            else:
                src_indices.append(resolve_direct_index(e))

        if pos_indices is None:
            raise ValueError(f"Index array {pos_name!r} must be used")

        return GatherSpec(
            out_rank=out_rank,
            axis_names=axis_names,
            src_name=src_name,
            src_indices=src_indices,
            pos_name=pos_name,
            pos_indices=pos_indices,
        )

    @staticmethod
    def parse(s: str, arr_names: Sequence[str]) -> "GatherSpec":
        return GatherSpec.resolve(parse.Equation.parse(s), list(arr_names))


if TYPE_CHECKING:
    ScatterReduce = Optional[Literal["add", "multiply"]]


@dataclass
class ScatterSpec:
    dst_indices: list[IdxExpr]
    src_indices: list[IdxExpr]
    pos_indices: list[IdxExpr]

    axis_names: list[str]
    dst_name: str
    src_name: str
    pos_name: str

    reduce: ScatterReduce

    def __post_init__(self) -> None:
        assert sum(1 for i in self.dst_indices if isinstance(i, PosDataIdx)) == 1
        assert all(
            isinstance(i, OutAxisIdx) or isinstance(i, IntLitIdx)
            for indices in [self.src_indices, self.pos_indices]
            for i in indices
        )
        assert all(
            i.axis < len(self.axis_names)
            for indices in [self.dst_indices, self.src_indices, self.pos_indices]
            for i in indices
            if isinstance(i, OutAxisIdx)
        )
        assert all(parse.is_ident(n) for n in self.axis_names)
        assert parse.is_ident(self.dst_name)
        assert parse.is_ident(self.src_name)
        assert parse.is_ident(self.pos_name)
        assert is_unique(self.axis_names + [self.dst_name, self.src_name, self.pos_name])

    def render(self, out: list[str]) -> None:
        renderer = IdxRenderer(axis_names=self.axis_names, pos_name=self.pos_name, pos_indices=self.pos_indices)
        out.append(self.dst_name)
        renderer.render_indices(self.dst_indices, out)
        out.append(" <- ")
        out.append(self.src_name)
        renderer.render_indices(self.src_indices, out)

    def __str__(self) -> str:
        out: list[str] = []
        self.render(out)
        return "".join(out)

    def __repr__(self) -> str:
        return f"ScatterSpec({self!s}, {[self.dst_name, self.src_name, self.pos_name]!r})"

    @staticmethod
    def resolve(equation: parse.Equation, arr_names: list[str], reduce: ScatterReduce) -> "ScatterSpec":
        if equation.kind != "<-":
            raise ValueError(f"Scatter formula must be written with '<-' operator; got: {equation.kind!r}")

        if len(arr_names) != 3:
            raise ValueError(f"Expected exactly three input array names for scatter; got: {arr_names!r}")

        if not is_unique(arr_names):
            raise ValueError(f"Input array names must be unique; got: {arr_names!r}")

        lhs = equation.lhs
        if not isinstance(lhs, parse.IdentExpr):
            raise ValueError(f"Invalid left hand side; expected array access expression, got: {lhs}")
        dst_name = lhs.name
        if dst_name not in arr_names:
            raise ValueError(f"Invalid 'dst' array name; expected one of {arr_names!r}, got: {dst_name!r}")

        rhs = equation.rhs
        if not isinstance(rhs, parse.IdentExpr):
            raise ValueError(f"Invalid right hand side; expected array access expression, got: {rhs}")
        src_name = rhs.name
        if src_name not in arr_names:
            raise ValueError(f"Invalid 'src' array name; expected one of {arr_names!r}, got: {src_name!r}")

        pos_name = next(n for n in arr_names if n not in [dst_name, src_name])

        axis_names: list[str] = []
        axis_ids: dict[str, int] = {}

        def resolve_direct_index(err_ctx: str, e: parse.Expr) -> Union[OutAxisIdx, IntLitIdx]:
            if isinstance(e, parse.IdentExpr) and e.idx == []:
                if e.name == pos_name:
                    raise ValueError(f"Scatter index array {pos_name!r} cannot appear in indices for {err_ctx}")
                axis_idx = axis_ids.get(e.name)
                if axis_idx is None:
                    axis_idx = len(axis_names)
                    axis_names.append(e.name)
                    axis_ids[e.name] = axis_idx
                return OutAxisIdx(axis_idx)
            elif isinstance(e, parse.IntExpr):
                return IntLitIdx(e.value)
            else:
                raise ValueError(f"Invalid index expression in {err_ctx}: {e}")

        dst_indices: list[IdxExpr] = []
        pos_indices: Optional[list[IdxExpr]] = None
        for i, e in enumerate(lhs.idx):
            if isinstance(e, parse.IdentExpr) and e.name == pos_name:
                dst_indices.append(PosDataIdx())
                if pos_indices is not None:
                    raise ValueError(f"Scatter index array {pos_name!r} must be used exactly once")
                pos_indices = [resolve_direct_index(repr(pos_name), e2) for e2 in e.idx]
            else:
                dst_indices.append(resolve_direct_index(repr(dst_name), e))

        if pos_indices is None:
            raise ValueError(f"Scatter index array {pos_name!r} must be used")

        src_indices: list[IdxExpr] = [resolve_direct_index(repr(src_name), e) for e in rhs.idx]

        all_axes = set(range(len(axis_names)))
        dst_axes = set(i.axis for i in dst_indices if isinstance(i, OutAxisIdx))
        pos_axes = set(i.axis for i in pos_indices if isinstance(i, OutAxisIdx))

        missing_dst = all_axes - dst_axes
        if len(missing_dst) > 1:
            raise ValueError(
                f"At most one index variable can be missing from indices used to directly access {dst_name!r} array; "
                + f"found {len(missing_dst)} missing indices: {[axis_names[i] for i in missing_dst]!r}"
            )

        missing_lhs = all_axes - dst_axes - pos_axes
        if len(missing_lhs) > 0:
            raise ValueError(
                f"All index variables must be used on left hand side of scatter formula; "
                + f"found {len(missing_lhs)} missing indices: {[axis_names[i] for i in missing_lhs]!r}"
            )

        return ScatterSpec(
            dst_indices=dst_indices,
            src_indices=src_indices,
            pos_indices=pos_indices,
            axis_names=axis_names,
            dst_name=dst_name,
            src_name=src_name,
            pos_name=pos_name,
            reduce=reduce,
        )

    @staticmethod
    def parse(s: str, arr_names: Sequence[str], reduce: ScatterReduce = None) -> "ScatterSpec":
        return ScatterSpec.resolve(parse.Equation.parse(s), list(arr_names), reduce)



# Cursed import hack file
# rust_circuit/python/rust_circuit is a legit python package called rust_circuit. When you `maturin dev`, rust is compiled and put into this folder, next to this file
# If you installed rust_circuit seperately, and have cloned this repo without compiling
# then you might accidentally import this version
# this script imports another installed compiled version of rust_circuit if it's installed
import os
import site

installed_rust_circuit_path = site.getsitepackages()[0] + "/rust_circuit"
if os.path.exists(installed_rust_circuit_path):
    import warnings

    warnings.warn("trying to rust_circuit local version without compiled rust. Using another installed version.")
    import importlib.util
    import sys

    rust_circuit_files = os.listdir(installed_rust_circuit_path)
    rust_so_file = [x for x in rust_circuit_files if x.startswith("_rust.cpython")][0]
    spec = importlib.util.spec_from_file_location("_rust", installed_rust_circuit_path + "/" + rust_so_file)
    foo = importlib.util.module_from_spec(spec)  # type: ignore
    print(foo)
    sys.modules["_rust"] = foo
    from _rust import *  # type: ignore
else:
    raise ImportError("rust_circuit is not compiled, run `maturin dev` to compile and install")



from functools import partial
from typing import Dict, Iterable, List, Literal, Optional, Sequence, TypeVar, Union

import torch

import interp.circuit.testing.topdown_circuit as td
import rust_circuit as rc

T = TypeVar("T")


def dedup_with_order(seq: Iterable[T]) -> List[T]:
    seen = set()
    out: List[T] = []
    for x in seq:
        if x in seen:
            continue
        else:
            seen.add(x)
            out.append(x)

    return out


def generate_with_shape(shape) -> torch.Tensor:
    return torch.randn(*shape, dtype=torch.float64)


def make_discrete_var(values_shape, name, batch_size, values_between_zero_one=False, group=None):
    """
    Values are sampled from random normal if values_between_zero_one=False,
    otherwise values are set to be in (0,1)
    """
    sample_shape = (batch_size,) + values_shape
    values = generate_with_shape(sample_shape)
    if values_between_zero_one:
        values = values ** 2
        values /= values.sum()

    out = rc.DiscreteVar(rc.Array(values), probs_and_group=group, name=name)
    return out, values, out.probs_and_group.evaluate()


st_Circuit = partial(td.st_Circuit, rust=True)



from __future__ import annotations

import abc
import inspect
import uuid
from typing import ClassVar, Optional, Tuple

import torch

from . import _rust as rc


class GeneralFunctionSpecBase(metaclass=abc.ABCMeta):
    __hash_bytes: ClassVar[bytes]
    """
    Inherit from this base class in order to implement an arbitrary new GeneralFunctionSpec.

    See docs for `get_shape_info`, GeneralFunctionShapeInfo, and `function`.
    """

    def __init_subclass__(cls):
        b = bytearray()
        for _, x in inspect.getmembers(cls, inspect.ismethod):
            if hasattr(x, "__code__"):
                b.extend(x.__code__.co_code)
                b.extend(uuid.UUID("6739c5c8-a9ab-421d-a06c-38f356bed339").bytes)
        b.extend(uuid.UUID("c5b1184c-76ae-430a-87cc-d6afc90569ce").bytes)
        b.extend((cls.__module__ + "." + cls.__name__).encode())
        cls.__hash_bytes = bytes(b)

    @classmethod
    def new(cls, *xs: rc.Circuit, name: Optional[str] = None):
        """This should be overridden if your generalfunction has non-Circuit arguments"""
        return rc.GeneralFunction(*xs, spec=cls(), name=name)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def path(self) -> Optional[str]:
        """The path of this spec relative to RRFS_DIR if this is stored in rrfs.
        Should be of the form /dir/sub_dir/.../file.py:MyCustomSpec"""
        return None

    def compute_hash_bytes(self) -> bytes:
        """What parts of self should be used for hashing & equality?

        should be implemented if you store data on `self`"""
        return self.__hash_bytes

    @abc.abstractmethod
    def function(self, *tensors: torch.Tensor) -> torch.Tensor:
        """run the actual function on tensors - these tensors shapes correspond to the shapes in ``get_shape_info``"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_shape_info(self, *shapes: rc.Shape) -> rc.GeneralFunctionShapeInfo:
        """This should error (exception) if the shapes are invalid and otherwise return GeneralFunctionShapeInfo"""
        raise NotImplementedError

    def get_device_dtype_override(self, *device_dtypes: rc.TorchDeviceDtypeOp) -> Optional[rc.TorchDeviceDtypeOp]:
        """Use this to check the validity of input dtypes/devices and set output device/dtype. If this returns None, the output inherits device/dtype from children.
        This can error to signal that inputs have invalid dtypes/devices"""
        return None
